"""Implements subcommands for ``ymp make`` and ``ymp submit``"""

import functools
import logging
import os
import shutil
import sys

import click

import ymp
from ymp.cli.shared_options import command, nohup_option
from ymp.common import Cache
from ymp.exceptions import YmpException, YmpStageError
from ymp.stage import StageStack

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


DEBUG_LOGFILE_NAME = os.environ.get("YMP_DEBUG_EXPAND")
if DEBUG_LOGFILE_NAME:
    import time
    start_time = time.time()
    if DEBUG_LOGFILE_NAME == "stderr":
        DEBUG_LOGFILE = sys.stderr
    else:
        DEBUG_LOGFILE = open(DEBUG_LOGFILE_NAME, "a")


def debug(msg, *args, **kwargs):
    if DEBUG_LOGFILE_NAME:
        tim = (time.time() - start_time)
        msg = "{:4.4f}: " + msg
        DEBUG_LOGFILE.write(msg.format(tim, *args, **kwargs) + '\n')


debug("started")


class TargetParam(click.ParamType):
    """Handles tab expansion for build targets"""

    @classmethod
    def complete(cls, _ctx, incomplete):
        """Try to complete incomplete command

        This is executed on tab or tab-tab from the shell

        Args:
          ctx: click context object
          incomplete: last word in command line up until cursor

        Returns:
          list of words incomplete can be completed to
        """
        result: list = []

        stack, _, tocomplete = incomplete.rpartition(".")
        debug("complete(stack={},incomplete={})", stack, tocomplete)

        if not stack:
            cfg = ymp.get_config()
            options = cfg.projects.keys()
            result += (o for o in options if o.startswith(tocomplete))
            result += (o + "." for o in options if o.startswith(tocomplete))
        else:
            from ymp.stage import StageStack
            try:
                stackobj = StageStack.get(stack)
            except YmpStageError as e:
                debug(e.format_message())
                return []
            debug("stacko = {}", repr(stack))
            options = stackobj.complete(tocomplete)
            debug("options = {}", options)
            # reduce items sharing prefix before "_"
            prefixes = {}
            for option in options:
                prefix = option.split("_", 1)[0]
                group = prefixes.setdefault(prefix, [])
                group.append(option)
            if len(prefixes) == 1:
                extensions = options
            else:
                extensions = []
                for prefix, group in prefixes.items():
                    if len(group) > 1:
                        extensions.append(prefix + "_")
                    else:
                        extensions.append(group[0])
            result += ('.'.join((stack, ext)) for ext in extensions)
            result += ('.'.join((stack, ext))+"." for ext in extensions
                       if not ext[-1] == "_")

        debug("res={}", result)
        return result


def snake_params(func):
    """Default parameters for subcommands launching Snakemake"""
    @click.argument(
        "targets", nargs=-1, metavar="TARGET_FILES",
        type=TargetParam()
    )
    @click.option(
        "--dryrun", "-n", default=False, is_flag=True,
        help="Only show what would be done"
    )
    @click.option(
        "--printshellcmds", "-p", default=False, is_flag=True,
        help="Print shell commands to be executed on shell"
    )
    @click.option(
        "--keepgoing", "-k", default=False, is_flag=True,
        help="Don't stop after failed job"
    )
    @click.option(
        "--lock/--no-lock",
        help="Use/don't use locking to prevent clobbering of files"
        " by parallel instances of YMP running"
    )
    @click.option(
        "--rerun-incomplete", "--ri", 'force_incomplete', is_flag=True,
        help="Re-run jobs left incomplete in last run"
    )
    @click.option(
        "--forceall", "-F", is_flag=True, default=False,
        help="Force rebuilding of all stages leading to target"
    )
    @click.option(
        "--force", "-f", "forcetargets", is_flag=True, default=False,
        help="Force rebuilding of target"
    )
    @click.option(
        "--notemp", is_flag=True, default=False,
        help="Do not remove temporary files"
    )
    @nohup_option
    @functools.wraps(func)
    def decorated(*args, **kwargs):  # pylint: disable=missing-docstring
        return func(*args, **kwargs)
    return decorated


class YmpConfigNotFound(YmpException):
    """
    Exception raised by YMP if no config was found in current path
    """
    pass


def start_snakemake(kwargs):
    """Execute Snakemake with given parameters and targets

    Fixes paths of kwargs['targets'] to be relative to YMP root.
    """
    cfg = ymp.get_config()
    if not cfg.projects:
        log.warning("No projects configured. Are you in the right folder?")
        log.warning("  Config files loaded:")
        for fname in cfg.conffiles:
            log.warning("    - %s", fname)

    root_path = cfg.root
    cur_path = os.path.abspath(os.getcwd())
    if not cur_path.startswith(root_path):
        raise YmpException("internal error - CWD moved out of YMP root?!")
    cur_path = cur_path[len(root_path):]

    # translate renamed arguments to snakemake synopsis
    arg_map = {
        'immediate': 'immediate_submit',
        'wrapper': 'jobscript',
        'scriptname': 'jobname',
        'cluster_cores': 'nodes',
        'snake_config': 'config',
        'drmaa': None,
        'sync': None,
        'sync_arg': None,
        'command': None,
        'args': None,
        'nohup': None
    }
    kwargs = {arg_map.get(key, key): value
              for key, value in kwargs.items()
              if arg_map.get(key, key) is not None}
    kwargs['workdir'] = root_path

    # our debug flag sets a new excepthoook handler, to we use this
    # to decide whether snakemake should run in debug mode
    if sys.excepthook.__module__ != "sys":
        log.warning(
            "Custom excepthook detected. Having Snakemake open stdin "
            "inside of run: blocks")
        kwargs['debug'] = True

    # map our logging level to snakemake logging level
    if log.getEffectiveLevel() > logging.WARNING:
        kwargs['quiet'] = True
    if log.getEffectiveLevel() < logging.WARNING:
        kwargs['verbose'] = True
    kwargs['use_conda'] = True
    if 'targets' in kwargs:
        if cur_path:
            kwargs['targets'] = [os.path.join(cur_path, t)
                                 for t in kwargs['targets']]
        else:
            targets = []
            for t in kwargs['targets']:
                try:
                    stack = StageStack.get(t)
                    targets.append(os.path.join(t, 'all_targets.stamp'))
                except YmpStageError as e:
                    #log.exception("asd")
                    targets.append(t)
            kwargs['targets'] = targets

    log.debug("Running snakemake.snakemake with args: %s", kwargs)

    # A snakemake workflow was created above to resolve the
    # stage stack. Unload it so things run correctly from within
    # snakemake.
    cfg.unload()

    import snakemake
    return snakemake.snakemake(ymp._snakefile, **kwargs)


@command()
@snake_params
@click.option(
    "--cores", "-j", default=1, metavar="CORES",
    help="The number of parallel threads used for scheduling jobs"
)
@click.option(
    "--dag", "printdag", default=False, is_flag=True,
    help="Print the Snakemake execution DAG and exit"
)
@click.option(
    "--rulegraph", "printrulegraph", default=False, is_flag=True,
    help="Print the Snakemake rule graph and exit"
)
@click.option(
    "--debug-dag", default=False, is_flag=True,
    help="Show candidates and selections made while the rule execution graph "
    "is being built"
)
@click.option(
    "--debug", default=False, is_flag=True,
    help="Set the Snakemake debug flag"
)
def make(**kwargs):
    "Build target(s) locally"
    rval = start_snakemake(kwargs)
    if not rval:
        sys.exit(1)


@command()
@snake_params
@click.option(
    "--profile", "-P", metavar="NAME",
    help="Select cluster config profile to use. Overrides cluster.profile "
    "setting from config."
)
@click.option(
    "--snake-config", "-c", metavar="FILE",
    help="Provide snakemake cluster config file"
)
@click.option(
    "--drmaa", "-d", is_flag=True,
    help="Use DRMAA to submit jobs to cluster. Note: Make sure you have "
    "a working DRMAA library. Set DRMAA_LIBRAY_PATH if necessary."
)
@click.option(
    "--sync", "-s", is_flag=True,
    help="Use synchronous cluster submission, keeping the submit command "
    "running until the job has completed. Adds qsub_sync_arg to cluster "
    "command"
)
@click.option(
    "--immediate", "-i", is_flag=True,
    help="Use immediate submission, submitting all jobs to the cluster "
    "at once."
)
@click.option(
    "--command", metavar="CMD",
    help="Use CMD to submit job script to the cluster"
)
@click.option(
    "--wrapper", metavar="CMD",
    help="Use CMD as script submitted to the cluster. See Snakemake "
    "documentation for more information."
)
@click.option(
    "--max-jobs-per-second", metavar="N",
    help="Limit the number of jobs submitted per second"
)
@click.option(
    "--latency-wait", "-l", metavar="T",
    help="Time in seconds to wait after job completed until files are "
    "expected to have appeared in local file system view. On NFS, this "
    "time is governed by the acdirmax mount option, which defaults to "
    "60 seconds."
)
@click.option(
    "--cluster-cores", "-J", type=int, metavar="N",
    help="Limit the maximum number of cores used by jobs submitted at a time"
)
@click.option(
    "--cores", "-j", default=16, metavar="N",
    help="Number of local threads to use"
)
@click.option(
    "--args", "args", metavar="ARGS",
    help="Additional arguments passed to cluster submission command. "
    "Note: Make sure the first character of the argument is not '-', "
    "prefix with ' ' as necessary."
)
@click.option(
    "--scriptname", metavar="NAME",
    help="Set the name template used for submitted jobs"
)
def submit(profile, **kwargs):
    """Build target(s) on cluster

    The parameters for cluster execution are drawn from layered profiles. YMP
    includes base profiles for the "torque" and "slurm" cluster engines.

    """
    cfg = ymp.get_config()

    # The cluster config uses profiles, which are assembled by layering
    # the default profile, the selected profile and additional command
    # line parameters. The selected profile is either specified via
    # "
    config = cfg.cluster.profiles.default
    profile_name = profile or cfg.cluster.profile
    if profile_name:
        config.add_layer(profile_name,
                         cfg.cluster.profiles[profile_name])
    cli_params = {key: value
                  for key, value in kwargs.items()
                  if value is not None}
    config.add_layer("<shell arguments>",
                     cli_params)

    # prepare cluster command
    if config.command is None:
        raise click.UsageError("""
        No cluster submission command configured.
        Please check the manual on how to configure YMP for your cluster"
        """)
    cmd = config.command.split() + config.args.values()
    if config.drmaa:
        param = 'drmaa'
        cmd[0] = ''
    elif config.sync:
        param = 'cluster_sync'
        cmd.append(config.sync_arg)
    else:
        param = 'cluster'

    if cmd[0] and not shutil.which(cmd[0]):
        raise click.UsageError(f"""
        The configured cluster submission command '{cmd[0]}' is does not
        exist or is not executable. Please check your cluster configuration.
        """)

    config.add_layer("<computed>", {param: cfg.expand(" ".join(cmd))})

    rval = start_snakemake(config)
    if not rval:
        sys.exit(1)



from ymp.helpers import odict, update_dict


def test_ordered_dict_maker():
    data = odict[
        'B': 2,
        'A': 1,
    ]
    assert data['A'] == 1
    assert data['B'] == 2
    for n, m in zip(data, ('B', 'A')):
        assert n == m


def test_update_dict():
    data = {'A': 1, 'B': 2, 'D': {'E': 3, 'F': 4}}

    assert update_dict(data, None) == data

    new_data = update_dict(data, {'C': 3})
    assert new_data['C'] == 3
    del new_data['C']
    assert new_data == data

    new_data = update_dict(data, {'D': {'E': 5, 'G': 6, 'H': {'I': 7}}})
    assert new_data['D']['E'] == 5
    assert new_data['D']['G'] == 6
    assert new_data['D']['H']['I'] == 7



import atexit
import glob
import logging
import os

from xdg import XDG_CACHE_HOME

import ymp.yaml
from ymp.common import AttrDict, Cache, MkdirDict, parse_number
from ymp.env import CondaPathExpander
from ymp.exceptions import YmpSystemError
from ymp.projects import Project
from ymp.references import Reference
from ymp.snakemake import \
    BaseExpander, \
    ColonExpander, \
    DefaultExpander, \
    ExpandableWorkflow, \
    InheritanceExpander, \
    RecursiveExpander, \
    SnakemakeExpander
from ymp.stage import StageExpander
from ymp.string import PartialFormatter

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class ConfigExpander(ColonExpander):
    def __init__(self, config_mgr):
        super().__init__()
        self.config_mgr = config_mgr

    def expands_field(self, field):
        return field not in 'func'

    class Formatter(ColonExpander.Formatter, PartialFormatter):
        def get_value(self, field_name, args, kwargs):
            cfg = self.expander.config_mgr

            # try to resolve variable as property of the config_mgr
            if hasattr(cfg, field_name):
                return getattr(cfg, field_name)
            return super().get_value(field_name, args, kwargs)


class OverrideExpander(BaseExpander):
    """
    Apply rule attribute overrides from ymp.yml config

    Example:
        Set the ``wordsize`` parameter in the `bmtagger_bitmask` rule to
        12:

        .. code-block:: yaml
           :caption: ymp.yml

           overrides:
             rules:
               bmtagger_bitmask:
                 params:
                   wordsize: 12
    """
    def __init__(self, cfgmgr):
        if 'overrides' not in cfgmgr._config:
            return
        self.rule_overrides = cfgmgr._config['overrides'].get('rules', {})
        super().__init__()

    def expand(self, rule, ruleinfo, **kwargs):
        overrides = self.rule_overrides.get(rule.name, {})
        for attr_name, values in overrides.items():
            attr = getattr(ruleinfo, attr_name)[1]
            for val_name, value in values.items():
                log.debug("Overriding {}.{}={} in {} with {}".format(
                    attr_name, val_name, attr[val_name], rule.name, value))
                attr[val_name] = value


class ConfigMgr(object):
    """Manages workflow configuration

    This is a singleton object of which only one instance should be around
    at a given time. It is available in the rules files as ``icfg`` and
    via `ymp.get_config()` elsewhere.

    ConfigMgr loads and maintains the workflow configuration as given
    in the ``ymp.yml`` files located in the workflow root directory,
    the user config folder (``~/.ymp``) and the installation ``etc``
    folder.
    """
    KEY_PROJECTS = 'projects'
    KEY_REFERENCES = 'references'
    CONF_FNAME = 'ymp.yml'
    CONF_DEFAULT_FNAME = ymp._defaults_file
    CONF_USER_FNAME = os.path.expanduser("~/.ymp/ymp.yml")
    RULE_MAIN_FNAME = ymp._snakefile

    __instance = None

    @classmethod
    def find_config(cls):
        """Locates ymp config files and ymp root

        The root ymp work dir is determined as the first (parent)
        directory containing a file named ``ConfigMgr.CONF_FNAME``
        (default ``ymp.yml``).

        The stack of config files comprises 1. the default config
        ``ConfigMgr.CONF_DEFAULT_FNAME`` (``etc/defaults.yml`` in the
        ymp package directory), 2. the user config
        ``ConfigMgr.CONF_USER_FNAME`` (``~/.ymp/ymp.yml``) and 3. the
        ``yml.yml`` in the ymp root.

        Returns:
          root: Root working directory
          conffiles: list of active configuration files
        """
        # always include defaults
        conffiles = [cls.CONF_DEFAULT_FNAME]

        # include user config if present
        if os.path.exists(cls.CONF_USER_FNAME):
            conffiles.append(cls.CONF_USER_FNAME)

        # try to find an ymp.yml in CWD and upwards
        filename = cls.CONF_FNAME
        log.debug("Locating '%s'", filename)
        curpath = os.path.abspath(os.getcwd())
        while not os.path.exists(os.path.join(curpath, filename)):
            log.debug("  not in '%s'", curpath)
            curpath, removed = os.path.split(curpath)
            if not removed:
                break
        if os.path.exists(os.path.join(curpath, filename)):
            root = curpath
            log.debug("  Found '%s' in '%s'", filename, curpath)
            conffiles.append(os.path.join(root, cls.CONF_FNAME))
        else:
            root = os.path.abspath(os.getcwd())
            log.debug("  No '%s' found; using %s as root", filename, root)

        return root, conffiles

    @classmethod
    def instance(cls):
        """Returns the active Ymp ConfigMgr instance

        """
        if cls.__instance is None:
            cls.__instance = cls(*cls.find_config())
        return cls.__instance

    @classmethod
    def activate(cls):
        ExpandableWorkflow.activate()

    @classmethod
    def unload(cls):
        log.debug("Unloading ConfigMgr")
        ExpandableWorkflow.clear()
        if cls.__instance:
            cls.__instance.cache.close()
        cls.__instance = None
        from ymp.stage import Stage, StageStack
        StageStack.stacks = {}
        Stage.active = None

    def __init__(self, root, conffiles):
        log.debug("Inizializing ConfigMgr")
        self.root = root
        self.conffiles = conffiles

        if os.path.dirname(conffiles[-1]) == root:
            self.cachedir = os.path.join(self.root, ".ymp")
        else:
            self.cachedir = os.path.join(XDG_CACHE_HOME, "ymp")

        self._config = ymp.yaml.load(conffiles)
        self.cache = cache = Cache(self.cachedir)

        # lazy filled by accessors
        self._snakefiles = None

        self.projects = cache.get_cache(
            "projects",
            itemloadfunc=Project,
            itemdata=self._config.get(self.KEY_PROJECTS) or {},
            dependfiles=conffiles
        )

        self.references = cache.get_cache(
            "references",
            itemloadfunc=Reference,
            itemdata=self._config.get(self.KEY_REFERENCES) or {},
            dependfiles=conffiles
        )

        self._workflow = ExpandableWorkflow.register_expanders(
            SnakemakeExpander(),
            RecursiveExpander(),
            CondaPathExpander(self),
            StageExpander(),
            ConfigExpander(self),
            OverrideExpander(self),
            DefaultExpander(params=([], {
                'mem': self.mem(),
                'walltime': self.limits.default_walltime
            })),
            InheritanceExpander(),
        )

    @property
    def ref(self):
        """
        Configure references
        """
        return self.references

    @property
    def pairnames(self):
        return self._config.pairnames

    @property
    def conda(self):
        return self._config.conda

    @property
    def dir(self):
        """
        Dictionary of relative paths of named YMP directories

        The directory paths are relative to the YMP root workdir.
        """
        return self._config.directories

    @property
    def absdir(self):
        """
        Dictionary of absolute paths of named YMP directories
        """
        return AttrDict({name: os.path.abspath(os.path.expanduser(value))
                         for name, value in self.dir.items()})

    @property
    def ensuredir(self):
        """
        Dictionary of absolute paths of named YMP directories

        Directories will be created on the fly as they are requested.
        """
        return MkdirDict({name: os.path.abspath(os.path.expanduser(value))
                         for name, value in self.dir.items()})

    @property
    def cluster(self):
        """
        The YMP cluster configuration.
        """
        return self._config.cluster

    @property
    def limits(self):
        """
        The YMP limits configuration.
        """
        return self._config.limits

    @property
    def snakefiles(self):
        """
        Snakefiles used under this config in parsing order
        """
        if not self._snakefiles:
            self._snakefiles = [
                fn
                for dn in (os.path.dirname(self.RULE_MAIN_FNAME),
                           self.absdir.rules)
                for fn in sorted(glob.glob(os.path.join(dn, "**", "*.rules"),
                                           recursive=True),
                                 key=lambda v: v.lower())
                if os.path.basename(fn)[0] != "."
            ]
        return self._snakefiles

    def expand(self, item, **kwargs):
        expander = ConfigExpander(self)
        res = expander.expand(None, item, kwargs)
        return res

    def mem(self, base="0", per_thread=None, unit="m"):
        """Clamp memory to configuration limits

        Params:
           base:       base memory requested
           per_thread: additional mem required per allocated thread
           unit:       output unit (b, k, m, g, t)
        """
        mem = parse_number(base)
        max_mem = parse_number(self.limits.max_mem)
        if mem > max_mem:
            mem = max_mem
        min_mem = parse_number(self.limits.min_mem)
        if mem < min_mem:
            mem = min_mem

        div = parse_number("1"+unit)

        return int(mem / div)

    @property
    def shell(self):
        """The shell used by YMP

        Change by adding e.g. ``shell: /path/to/shell`` to ``ymp.yml``.
        """
        return self._config.shell

    @property
    def platform(self):
        """Name of current platform (macos or linux)"""
        if not (hasattr(self, '_platform') and self._platform):
            import platform
            system = platform.system()
            if system == "Darwin":
                self._platform = "macos"
            elif system == "Linux":
                self._platform = "linux"
            else:
                raise YmpSystemError(f"YMP does not support system '{system}'")
        return self._platform


atexit.register(ConfigMgr.unload)



"""
Various converters

"""

def BlastHit2GffFeature(blasthit):
    return gff.Feature(
        seqid = hit.sacc,
        source = 'BLAST',
        type = 'CDS',
        start = min(hit.sstart, hit.send),
        end = max(hit.sstart, hit.send),
        score = hit.evalue,
        strand = '+' if hit.sframe > 0 else '-',
        phase = '0',
        attributes = "ID={}_{}_{}".format(hit.sacc, hit.sstart, hit.send)
    )



import logging
import os
import re
from collections import Mapping, Sequence

import ymp
from ymp.common import ensure_list
from ymp.exceptions import YmpConfigError, YmpStageError
from ymp.stage import Stage
from ymp.util import is_fq, make_local_path

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class PandasTableBuilder(object):
    """Builds the data table describing each sample in a project

    This class implements loading and combining tabular data files
    as specified by the YAML configuration.

    Format:
     - string items are files
     - lists of files are concatenated top to bottom
     - dicts must have one "command" value:

       - 'join' contains a two-item list
         the two items are joined 'naturally' on shared headers
       - 'table' contains a list of one-item dicts
         dicts have form ``key:value[,value...]``
         a in-place table is created from the keys
         list-of-dict is necessary as dicts are unordered
       - 'paste' contains a list of tables pasted left to right
         tables pasted must be of equal length or length 1
     - if a value is a valid path relative to the csv/tsv/xls file's
       location, it is expanded to a path relative to CWD

    Example:
     .. code-block:: yaml

      - top.csv
      - join:
        - excel.xslx%left.csv
        - right.tsv
      - table:
        - sample: s1,s2,s3
        - fq1: s1.1.fq, s2.1.fq, s3.1.fq
        - fq2: s1.2.fq, s2.2.fq, s3.2.fq

    """
    """Constructs the project data table from the yaml definition
    """
    def __init__(self):
        import pandas
        from pandas.core.reshape.merge import MergeError
        self.pd = pandas
        self.MergeError = MergeError
        self.files: List[str] = []

    def load_data(self, cfg):

        if isinstance(cfg, str):
            return self._load_file(cfg)
        if isinstance(cfg, Sequence):
            return self._rowbind(cfg)
        if isinstance(cfg, Mapping):
            if 'join' in cfg:
                return self._join(cfg['join'])
            if 'paste' in cfg:
                return self._paste(cfg['paste'])
            if 'table' in cfg:
                return self._table(cfg['table'])
        raise YmpConfigError(cfg, "Unrecognized statement in data config")

    def _load_file(self, cfg):
        try:
            data = self.pd.read_csv(
                cfg, sep=None, engine='python', dtype='str'
            )
        except FileNotFoundError:
            parts = cfg.split('%', maxsplit=1)
            try:
                data = self.pd.read_excel(
                    parts[0], parts[1] if len(parts) > 1 else 0)
            except ImportError:
                raise YmpConfigError(
                    cfg,
                    "Could not load specified data file."
                    " If this is an Excel file, you might need"
                    " to install 'xlrd'."
                )
        # prefix fq files with name of config file's directory
        rdir = os.path.dirname(cfg)
        data = data.applymap(
            lambda s: os.path.join(rdir, s)
            if is_fq(s) and os.path.exists(os.path.join(rdir, s))
            else s
        )
        self.files.append(cfg)
        return data

    def _rowbind(self, cfg):
        tables = list(map(self.load_data, cfg))
        return self.pd.concat(tables, ignore_index=True)

    def _join(self, cfg):
        tables = list(map(self.load_data, cfg))
        try:
            return self.pd.merge(*tables)
        except self.MergeError as e:
            raise YmpConfigError(
                cfg,
                "Failed to `join` configured data.\n"
                "Joined table indices:\n{}\n\n"
                "".format("\n".join(", ".join(table.columns.tolist())
                                    for table in tables)),
                exc=e)

    def _paste(self, cfg):
        tables = list(map(self.load_data, cfg))
        manyrow = [table for table in tables if len(table) > 1]
        if manyrow:
            nrows = len(manyrow[0])
            if any(len(table) != nrows for table in manyrow[1:]):
                raise YmpConfigError(
                    cfg,
                    "Failed to `paste` configured data. "
                    "Row counts differ and are not 1."
                    "Row counts: {}\n"
                    "".format(", ".join(str(len(table))
                                        for table in manyrow)))
            tables = [
                table if len(table) > 1
                else self.pd.concat([table]*nrows, ignore_index=True)
                for table in tables
            ]
        return self.pd.concat(tables, axis=1)

    def _table(self, cfg):
        return self.pd.DataFrame.from_dict({
            key: value.split(',')
            for row in cfg
            for key, value in row.items()
        })


class PandasProjectData(object):
    def __init__(self, cfg):
        import pandas
        self.pd = pandas
        table_builder = PandasTableBuilder()
        self.df = table_builder.load_data(cfg)

    def columns(self):
        return list(self.df.columns)

    def identifying_columns(self):
        column_frequencies = self.df.apply(self.pd.Series.nunique)
        log.debug("Column frequencies: {}".format(column_frequencies))
        nrows = self.df.shape[0]
        log.debug("Row count: {}".format(nrows))
        columns = self.df.columns[column_frequencies == nrows]
        return list(columns)

    def dump(self):
        return self.df.to_dict()

    def duplicate_rows(self, column):
        duplicated = self.df.duplicated(subset=[column], keep=False)
        values = self.df[duplicated][column]
        return list(values)

    def string_columns(self):
        cols = self.df.select_dtypes(include=['object'])
        # turn NaN into '' so they don't bother us later
        cols.fillna('', inplace=True)
        return list(cols)

    def rows(self, cols):
        yield from self.df[cols].itertuples()

    def get(self, idcol, row, col):
        return ensure_list(self.df[self.df[idcol] == row][col].values[0])

    def column(self, col):
        return list(self.df[col])

    def groupby_dedup(self, cols):
        skip = set()
        result = []
        df = self.df
        for g in cols:
            if g in skip:
                continue
            result.append(g)
            # get columns constant for current g
            rcols = set(df.columns[df.groupby(result).agg('nunique').eq(1).all()])
            # mark redundant columns
            skip |= rcols
        return result


class SQLiteProjectData(object):
    def __init__(self, cfg, name="data"):
        import sqlite3
        self.conn = sqlite3.connect(":memory:")
        self.name = name
        table_builder = PandasTableBuilder()
        table_builder.load_data(cfg).to_sql(name, self.conn, index=False)

    def query(self, *args):
        return self.conn.execute(*args)

    def __getstate__(self):
        return (self.name, self.dump())

    def __setstate__(self, state):
        import sqlite3
        self.conn = sqlite3.connect(":memory:")
        self.name = state[0]
        self.conn.executescript(state[1])

    @property
    def nrows(self):
        return self.query(
            'SELECT COUNT(*) from "{}"'.format(self.name)
        ).fetchone()[0]

    def columns(self):
        return [t[1] for t in self.query(
            'PRAGMA table_info("{}")'.format(self.name)
        )]

    def identifying_columns(self):
        nrows = self.nrows
        return [t for t in self.columns() if self.query(
            'SELECT COUNT(DISTINCT "{}") FROM "{}"'.format(t, self.name)
        ).fetchone()[0] == nrows]

    def dump(self):
        return "\n".join(self.conn.iterdump())

    def duplicate_rows(self, column):
        return [t[0] for t in self.query("""
        SELECT "{c}" FROM "{n}" WHERE "{c}" in (
                 SELECT "{c}" FROM "{n}" GROUP BY "{c}" having COUNT("{c}") > 1
        )
        """.format(c=column, n=self.name)
        )]

    def string_columns(self):
        return [t[1] for t in self.query(
            'PRAGMA table_info("{}")'.format(self.name)
        ) if t[2] == 'TEXT']

    def rows(self, col):
        return self.query(
            'SELECT {c} from "{n}"'.format(
                n=self.name,
                c="0," + ",".join('"{}"'.format(c) for c in col)
            )
        )

    def get(self, idcol, row, col):
        query = f'SELECT "{col}" from "{self.name}" where "{idcol}"=?'
        res = [row[0] for row in self.query(query, [row]).fetchall()]
        return res

    def column(self, col):
        if isinstance(col, str):
            col = [col]
        col = ",".join('"{}"'.format(c) for c in col)
        return [t[0] for t in self.query(
            'SELECT {c} from "{n}"'.format(n=self.name, c=col)
        )]

    def groupby_dedup(self, cols):
        skip = set()
        result = []
        for g in cols:
            if g in skip:
                continue
            result.append(g)
            q = 'SELECT {c} FROM "{n}" GROUP BY {g}'.format(
                n=self.name,
                g=",".join('"{}"'.format(g) for g in result),
                c=",".join(('COUNT(DISTINCT "{}")'.format(c) for c in cols))
            )
            counts = self.query(q).fetchall()
            unique = [all(x == 1 for x in n) for n in zip(*counts)]
            rcols = set(c for c, u in zip(cols, unique) if u)
            skip |= rcols
        return result


class Project(Stage):
    """Contains configuration for a source dataset to be processed"""
    KEY_DATA = 'data'
    KEY_IDCOL = 'id_col'
    KEY_READCOLS = 'read_cols'
    KEY_BCCOL = 'barcode_col'

    # SRR columns:
    # LibraryLayout_s PAIRED
    # LibrarySelection_s PCR | RANDOM

    RE_REMOTE = re.compile(r"^(?:https?|ftp|sftp)://(?:.*)")
    RE_SRR = re.compile(r"^[SED]RR[0-9]+$")
    RE_FILE = re.compile(r"^(?!http://).*(?:fq|fastq)(?:|\.gz)$")

    def __init__(self, project, cfg):
        # Fixme: put line in config here
        self.filename = "fn"
        self.lineno = 0
        # triggers early workflow load, breaking things... :/
        # super().__init__(project)
        self.project = project
        self.name = project
        self.altname = None
        self.cfg = cfg
        self.pairnames = ymp.get_config().pairnames
        self.fieldnames = None
        self._data = None
        self._source_cfg = None
        self._idcol = None
        self.bccol = cfg.get(self.KEY_BCCOL)
        self.outputs = set(("/{sample}.R1.fq.gz", "/{sample}.R2.fq.gz",
                            "/{:samples:}.R1.fq.gz", "/{:samples:}.R2.fq.gz"))
        self.inputs = set()

        if self.KEY_DATA not in self.cfg:
            raise YmpConfigError(
                self.cfg, "Missing key '{}'".format(self.KEY_DATA))

    def __repr__(self):
        return "{}(project={})".format(self.__class__.__name__, self.project)

    @property
    def defined_in(self):
        return self.cfg.get_files()

    @property
    def data(self):
        """Pandas dataframe of runs

        Lazy loading property, first call may take a while.
        """
        if self._data is None:
            self._data = SQLiteProjectData(self.cfg[self.KEY_DATA])

        return self._data

    @property
    def variables(self):
        return self.data.columns()

    def minimize_variables(self, groups):
        if not groups:
            groups = [self.idcol]
        if len(groups) > 1:
            groups = [g for g in groups if g != 'ALL']
        if len(groups) > 1:
            groups = self.data.groupby_dedup(groups)
        if len(groups) > 1:
            raise YmpStageError("multi-idx grouping not implemented")
        return groups

    def get_ids(self, groups, match_groups=None, match_value=None):
        if groups == ['ALL']:
            return 'ALL'
        if groups == match_groups:
            return match_value
        if match_groups and match_groups != ['ALL']:
            return self.data.get(match_groups[0], match_value, groups[0])
        else:
            return self.data.column(groups[0])

    def iter_samples(self, variables=None):
        if not variables:
            variables = self.variables
        return self.data.rows(variables)

    @property
    def runs(self):
        """Pandas dataframe index of runs

        Lazy loading property, first call may take a while.
        """
        return self.data.column(self.idcol)

    @property
    def idcol(self):
        if self._idcol is None:
            self._idcol = self.choose_id_column()
        return self._idcol

    @property
    def source_cfg(self):
        if self._source_cfg is None:
            self._source_cfg = self.choose_fq_columns()
        return self._source_cfg

    def choose_id_column(self):
        """Configures column to use as index on runs

        If explicitly configured via KEY_IDCOL, verifies that the column
        exists and that it is unique. Otherwise chooses the leftmost
        unique column in the data.
        """
        all_columns = self.data.columns()
        unique_columns = self.data.identifying_columns()

        if not unique_columns:
            raise YmpConfigError(
                self.cfg,
                "Project data has no column containing unique values for "
                "each row. At least one is needed to identify samples!"
            )

        if self.KEY_IDCOL in self.cfg:
            idcol = self.cfg[self.KEY_IDCOL]
            if idcol not in all_columns:
                raise YmpConfigError(self.cfg, key=self.KEY_IDCOL, msg=(
                    "Configured column not found in data. "
                    "Possible spelling error? Available columns: "
                    ", ".join(all_columns)
                    ))

            if idcol not in unique_columns:
                raise YmpConfigError(self.cfg, key=self.KEY_IDCOL, msg=(
                    "Configured id_col column '{}' is not unique.\n"
                    "Duplicated rows:\n {}\n"
                    "Unique columns: {}".format(
                        idcol, self.data.duplicate_rows(idcol), unique_columns
                    )
                ))
        else:
            idcol = unique_columns[0]
            log.info("Autoselected column %s=%s", self.KEY_IDCOL, idcol)

        return idcol

    def choose_fq_columns(self):
        """
        Configures the columns referencing the fastq sources
        """
        # get only columns containing string data
        string_cols = self.data.string_columns()

        # if barcode column specified, omit that
        if self.bccol:
            string_cols.remove(self.bccol)

        # if read columns specified, constrain to those
        read_cols = self.cfg.get(self.KEY_READCOLS)
        if read_cols:
            if isinstance(read_cols, str):
                read_cols = [read_cols]
            typo_cols = set(read_cols) - set(string_cols)
            if typo_cols:
                log.warning("%s=%s references invalid columns: %s",
                            self.KEY_READCOLS, read_cols, typo_cols)
                read_cols = [col for col in read_cols if col not in typo_cols]
        else:
            read_cols = string_cols

        if not read_cols:
            raise YmpConfigError(self.cfg, key=self.KEY_READCOLS, msg=(
                "No columns containing read files found"
            ))

        err = False
        source_config = {}
        for row in self.data.rows([self.idcol] + read_cols):
            cols = []
            for i, val in enumerate(row[2:]):
                if self.RE_FILE.match(val):
                    cols.append(("file", read_cols[i]))
                elif self.RE_REMOTE.match(val):
                    cols.append(("remote", read_cols[i]))
                elif self.RE_SRR.match(val):
                    cols.append(("srr", read_cols[i]))
            types = set(col[0] for col in cols)
            if not types:
                log.error("No data sources found in row %s.",
                          row[1])
                err = True
            elif len(types) > 1 or len(cols) > 2 or \
                 (cols[0] == 'srr' and len(cols) > 1):
                log.error("Ambiguous data sources found in row %s. "
                          "You may need to constrain the columns allowed "
                          "to contain read data using '%'.",
                          row[1], self.KEY_READCOLS)
                err = True
            elif len(cols) == 2:
                source_config[row[1]] = (cols[0][0], cols[0][1], cols[1][1])
            elif len(cols) == 1:
                source_config[row[1]] = (cols[0][0], cols[0][1], None)
            else:
                raise RuntimeError("this should not have happened")
        if err:
            raise YmpConfigError(self.cfg, msg=(
                "Failed to identify source data in project data config. "
                "See above log messages for details."
            ))

        return source_config

    def source_path(self, run, pair, nosplit=False):
        """Get path for FQ file for ``run`` and ``pair``"""
        source = self.source_cfg.get(run)
        cfg = ymp.get_config()
        if not source:
            raise YmpConfigError(self.cfg,
                                 "No run '{}' in source config".format(run))

        if isinstance(pair, str):
            pair = self.pairnames.index(pair)

        if self.bccol and not nosplit:
            barcode_file = self.data.get(self.idcol, run, self.bccol)[0]
            if barcode_file:
                return self.encode_barcode_path(barcode_file, run, pair)

        kind = source[0]
        if kind == 'srr':
            srr = self.data.get(self.idcol, run, source[1])[0]
            f = os.path.join(cfg.dir.scratch,
                             "SRR",
                             "{}_{}.fastq.gz".format(srr, pair+1))
            return f

        fq_col = source[pair+1]
        if not isinstance(fq_col, str):
            return (
                "Configuration Error: no source for sample {} and read {} "
                "found.".format(run, pair+1))

        fn = self.data.get(self.idcol, run, fq_col)[0]
        if kind == 'file':
            return fn

        if kind == 'remote':
            return make_local_path(cfg, fn)

        raise YmpConfigError(
            self.cfg,
            "Configuration Error: no source for sample {} and read {} found."
            "".format(run, pair+1))

    def encode_barcode_path(self, barcode_file, run, pair):
        if barcode_file:
            barcode_id = barcode_file.replace("_", "__").replace("/", "_%")
            return (
                "{project}.split_libraries/{barcodes}/{run}.{pair}.fq.gz"
                "".format(
                    project=self.project,
                    barcodes=barcode_id,
                    run=run,
                    pair=self.pairnames[pair])
            )

    def unsplit_path(self, barcode_id, pairname):
        barcode_file = barcode_id.replace("_%", "/").replace("__", "_")
        pair = self.pairnames.index(pairname)

        run = self.data.get(self.bccol, barcode_file, self.idcol)[0]
        source = self.source_path(run, pair, nosplit=True)
        return [barcode_file, source]

    def get_fq_names(self,
                     only_fwd=False, only_rev=False,
                     only_pe=False, only_se=False):
        """Get pipeline names of fq files"""

        if only_fwd and only_rev:  # pointless, but satisfiable
            return []
        if only_pe and only_se:  # equally pointless, zero result
            return []

        pairs = []
        if not only_rev:
            pairs += [0]
        if not only_fwd:
            pairs += [1]

        check_rev = only_pe or only_se

        def have_file(run, pair):
            return (isinstance(self.source_cfg[run][pair+1], str)
                    or self.source_cfg[run][0] == 'srr')

        return [
            "{}.{}".format(run, self.pairnames[pair])
            for run in self.runs
            for pair in pairs
            if have_file(run, pair)
            if not check_rev or have_file(run, 1) == only_pe
        ]

    @property
    def fq_names(self):
        "Names of all FastQ files"
        return self.get_fq_names()

    @property
    def pe_fq_names(self):
        "Names of paired end FastQ files"
        return self.get_fq_names(only_pe=True)

    @property
    def se_fq_names(self):
        "Names of single end FastQ files"
        return self.get_fq_names(only_se=True)

    @property
    def fwd_pe_fq_names(self):
        "Names of forward FastQ files part of pair"
        return self.get_fq_names(only_pe=True, only_fwd=True)

    @property
    def rev_pe_fq_names(self):
        "Names of reverse FastQ files part of pair"
        return self.get_fq_names(only_pe=True, only_rev=True)

    @property
    def fwd_fq_names(self):
        "Names of forward FastQ files (se and pe)"
        return self.get_fq_names(only_fwd=True)



import re

from itertools import product
from string import Formatter
from typing import List, Dict, Tuple, Union, Any, Set

import snakemake.utils


class FormattingError(AttributeError):
    def __init__(self, message: str, fieldname: str) -> None:
        super().__init__(message)
        self.attr = fieldname


class GetNameFormatter(Formatter):
    def get_names(self, pattern: str):
        for val in self.parse(pattern):
            if val[1] is not None:
                yield val[1]


class OverrideJoinFormatter(Formatter):
    """Formatter with overridable join method

    The default formatter joins all arguments with
    ``"".join(args)``. This class overrides :meth:`_vformat` with
    identical code, changing only that line to one that can be
    overridden by a derived class.
    """

    def _vformat(self,
                 format_string: str,
                 args: List,
                 kwargs: Dict,
                 used_args,
                 recursion_depth: int,
                 auto_arg_index: int=0) -> Tuple[Union[List[str], str], int]:
        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')
        result = []
        for literal_text, field_name, format_spec, conversion in \
                self.parse(format_string):

            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do
                #  the formatting

                # handle arg indexing when empty field_names are given.
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # given the field_name, find the object it references
                #  and the argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # expand the format spec, if needed
                format_spec, auto_arg_index = self._vformat(
                    format_spec, args, kwargs,
                    used_args, recursion_depth-1,
                    auto_arg_index=auto_arg_index)

                result.append(self.format_field(obj, format_spec))

        return self.join(result), auto_arg_index

    def join(self, args: List[str]) -> Union[List[str],str]:
        """
        Joins the expanded pieces of the template string to form the output.

        This function is equivalent to ``''.join(args)``. By overriding it,
        alternative methods can be implemented, e.g. to create a list of
        strings, each corresponding to a the cross product of the expanded
        variables.
        """
        return ''.join(args)


class ProductFormatter(OverrideJoinFormatter):
    """
    String Formatter that creates a list of strings each expanded using
    one point in the cartesian product of all replacement values.

    If none of the arguments evaluate to lists, the result is a string,
    otherwise it is a list.

    >>> ProductFormatter().format("{A} and {B}", A=[1,2], B=[3,4])
    "1 and 3"
    "1 and 4"
    "2 and 3"
    "2 and 4"
    """
    def join(self, args: List[Any]) -> Union[List[str], str]:
        # expand everything that isn't a string to a list
        args = [[item] if isinstance(item, str) else list(item)
                for item in args]
        # combine items into list corresponding to cartesian product
        res = [''.join(flat_args) for flat_args in product(*args)]

        if len(res) > 1:
            return res

        if res:
            return res[0]

        return ''

    def format_field(self, value, format_spec: str):
        if hasattr(value, '__iter__') and not isinstance(value, str):
            return (format(item) for item in value)
        return format(value, format_spec)


class RegexFormatter(Formatter):
    """
    String Formatter accepting a regular expression defining the format
    of the expanded tags.
    """
    def __init__(self, regex: Union[str, Any]) -> None:
        super().__init__()
        if (isinstance(regex, str)):
            self.regex = re.compile(regex)
        else:
            self.regex = regex

    def parse(self, format_string: str):
        """
        Parse format_string into tuples. Tuples contain
        literal_text: text to copy
        field_name: follwed by field name
        format_spec:
        conversion:
        """
        if format_string is None:
            return

        start = 0
        for match in self.regex.finditer(format_string):
            yield (format_string[start:match.start()],  # literal text
                   match.group('name'),                 # field name
                   '',                                  # format spec
                   None)                                # conversion
            start = match.end()

        # yield text at end of format_string
        yield (format_string[start:],
               None, None, None)

    def get_names(self, format_string: str) -> Set[str]:
        """Get set of field names in format_string)"""
        return set(match.group('name')
                   for match in self.regex.finditer(format_string))


class QuotedElementFormatter(snakemake.utils.SequenceFormatter):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.element_formatter = snakemake.utils.QuotedFormatter()


class PartialFormatter(Formatter):
    """
    Formats what it can and leaves the remainder untouched
    """

    def get_field(self, field_name, args, kwargs):
        try:
            val = super().get_field(field_name, args, kwargs)
            if type(val[0]).__name__ == "function":
                raise IndexError()
            return val
        except (KeyError, IndexError, TypeError):
            return getattr(self, "spec", "{{{}}}").format(field_name), None


def make_formatter(product=None, regex=None, partial=None, quoted=None):
    formatter = 1
    types: 'List[type]' = []
    class_name = ""
    class_dict = {}
    for arg, cls, name in (
            (product, ProductFormatter, 'Product'),
            (regex, RegexFormatter, 'Regex'),
            (partial, PartialFormatter, 'Partial'),
            (quoted, QuotedElementFormatter, 'QuotedElement'),
            (formatter, GetNameFormatter, 'Formatter')
             ):
        if arg is not None:
            types += [cls]
            class_name += name
            if not isinstance(arg, int):
                class_dict[name.lower()] = arg
    return type(class_name, tuple(types), {})(**class_dict)



AA = 'FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'
NU = 'TCAG'
B2N = { a:b for a,b in zip(NU, range(len(NU))) }
nuc2num = lambda seq: sum([
    len(NU) ** pos * B2N[nuc]
    for pos,nuc in enumerate(reversed(seq))
])
nuc2aa = lambda seq: ''.join([
    AA[nuc2num(codon)]
    for codon in zip(*[iter(seq)]*3)
])







import logging
import os
import shlex
import shutil

import py

import pytest

import ymp
import ymp.config

log = logging.getLogger(__name__)


# Add pytest options
# ==================

def pytest_addoption(parser):
    parser.addoption("--run-tools", action="store_true",
                     default=False, help="Skip tests running tools")
    parser.addoption("--cwd-save-dir", metavar="DIR",
                     default="test_failures",
                     help="""Tests needing local files are run in a temporary
                     directory. If a test fails, a copy of the directory is
                     made in this location.""")
    parser.addoption("--cwd-save-always", action="store_true",
                     default=False, help="Always save test CWD")


# Add pytest.mark.skip_tool
# =========================

def pytest_collection_modifyitems(config, items):
    if not config.getoption("--run-tools"):
        skip_tool = pytest.mark.skip(reason="Not running tools")
        for item in items:
            if "runs_tool" in item.keywords:
                item.add_marker(skip_tool)


# Allow executing tests in dir saved on error
# ===========================================

@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # set a report attribute for each phase of a call, which can
    # be "setup", "call", "teardown"
    setattr(item, "rep_" + rep.when, rep)


@pytest.fixture()
def saved_tmpdir(request, tmpdir):
    yield tmpdir
    if (
        request.config.getoption("--cwd-save-always")
        or not hasattr(request.node, 'rep_call')
        or request.node.rep_call.failed
    ):
        name_parts = request.node.name.replace("]", "").split("[")
        cwd_save_dir = request.config.getoption("--cwd-save-dir")
        destdir = py.path.local(cwd_save_dir).join(*name_parts)
        # delete conda / conda_archive before saving tmpdir
        for path in ("conda", "conda_archive"):
            delpath = tmpdir.join(path)
            if delpath.check(exists=1):
                delpath.remove(rec=True)
        if destdir.check(exists=1):
            destdir.remove(rec=True)
        destdir.dirpath().ensure_dir()
        tmpdir.move(destdir)
        log.error("Saved failed test data to %s", str(destdir))
    else:
        tmpdir.remove()


@pytest.fixture()
def saved_cwd(saved_tmpdir):
    with saved_tmpdir.as_cwd():
        yield saved_tmpdir


# Inject executables into PATH
# ==============================

@pytest.fixture()
def bin_dir(saved_tmpdir):
    binpath = os.path.join(saved_tmpdir, "bin")
    try:
        os.mkdir(binpath)
    except FileExistsError:
        if not os.path.isdir(binpath):
            raise
    path = os.environ['PATH']
    os.environ['PATH'] = ':'.join((binpath, path))
    yield binpath
    os.environ['PATH'] = path


class MockCmd(object):
    _calls = None

    def __init__(self, bin_dir, name, code=""):
        self.filename = os.path.join(bin_dir, name)
        self.logname = self.filename + "_cmd.log"
        content = "\n".join([
            '#!/bin/sh',
            'echo "$0 $@" >> "{}"'.format(self.logname),
            '\n'
        ])
        with open(self.filename, "w") as f:
            f.write(content)
            f.write(code)
        os.chmod(self.filename, 0o500)

    @property
    def calls(self):
        if not os.path.exists(self.logname):
            log.debug("%s is empty", self.logname)
            return []
        with open(self.logname) as r:
            data = r.read().splitlines()
        log.debug("%s:\n |  %s", self.logname, "\n | ".join(data))
        return data


@pytest.fixture
def mock_cmd(request, bin_dir):
    cmd, script = request.param
    print("###", cmd, script)
    yield MockCmd(bin_dir, cmd, script)


@pytest.fixture
def mock_conda(bin_dir):
    base_dir = os.path.dirname(bin_dir)
    yield MockCmd(bin_dir, "conda", "\n".join([
        'cmd=""',
        'while [ -n "$1" ]; do',
        '  case $1 in',
        '  --version)   echo conda 4.2; exit 0;;',
        '  --prefix|-p) shift; p="$1";;',
        '  --file|-f)   shift; f="$1";;'
        '  --json)      shift; j=Y;;'
        '  *)           cmd="$cmd $1";;',
        '  esac',
        '  shift',
        'done',
        'if echo "$cmd" |grep -q " create "; then',
        '  mkdir -p "$p"',
        'fi',
        'if [ x"$cmd" = x" env export" -a -n "$p" ]; then',
        '  echo "dependencies: [one, two]"',
        'fi',
        'if [ x"$cmd" = x" info" ]; then',
        '  echo \'{{"conda_prefix": "{}"}}\''.format(base_dir),
        'fi',
    ]))


# Activate this to get some profiling data while testing
# ======================================================

@pytest.fixture(scope="module")  # autouse=True)
def profiling():
    import yappi
    yappi.start()
    yield
    yappi.stop()
    profile = yappi.get_func_stats()
    profile.sort("subtime")
    with open("profile.txt", "w") as f:
        profile.print_all(out=f, columns={
            0: ("name", 120),
            1: ("ncall", 10),
            2: ("tsub", 8),
            3: ("ttot", 8),
            4: ("tavg", 8)})


# Provision CWD with files from data/<project>
# ============================================

@pytest.fixture()
def project(request):
    "Parametrizable project; defaults to 'toy'"
    return getattr(request, 'param', "toy")


@pytest.fixture()
def demo_dir(invoker, saved_cwd):
    invoker.call("init", "demo")
    return saved_cwd


@pytest.fixture()
def project_dir(request, project, saved_tmpdir):
    """Populated project directory

    parametrize `project` to get different project dirs
    """
    data_dir = py.path.local(__file__).dirpath('data', project)
    data_dir.copy(saved_tmpdir)
    log.info("Created project directory {}".format(saved_tmpdir))
    yield saved_tmpdir
    log.info("Tearing down project directory {}".format(saved_tmpdir))


@pytest.fixture()
def target(request, project_dir):
    with project_dir.as_cwd():
        log.info("Switched to directory {}".format(project_dir))
        ymp.get_config().unload()
        cfg = ymp.get_config()
        targets = [request.param.format(prj) for prj in cfg.projects]
        with open("target.txt", "w") as out:
            out.write("\n".join(targets))
        yield from targets


# Call into CLI
# =============

class Invoker(object):
    """Wrap invoking shell command

    Handles writing of out.log and cmd.sh as well as reloading ymp config
    on each call.
    """
    def __init__(self):
        from click.testing import CliRunner
        self.runner = CliRunner()
        from ymp.cli import main
        self.main = main
        self.initialized = False
        self.toclean = []

    def call(self, *args, standalone_mode=False, **kwargs):
        """Call into YMP CLI

        ``standalone_mode`` defaults to False so that exceptions are
        passed rather than caught.

        """
        if not self.initialized:
            # change path to USER ymp config (default ~/.ymp/ymp.yml)
            # so that settings there do not interfere with tests
            ymp.config.ConfigMgr.CONF_USER_FNAME = "ymp_user.yml"
            # force reload
            ymp.get_config().unload()

        if not os.path.exists("cmd.sh"):
            with open("cmd.sh", "w") as f:
                f.write("#!/bin/bash -x\n")

        argstr = " ".join(shlex.quote(arg) for arg in args)
        with open("cmd.sh", "w") as f:
            f.write(f"PATH={os.environ['PATH']} ymp {argstr} \"$@\"\n")

        result = self.runner.invoke(self.main, args, **kwargs,
                                    standalone_mode=standalone_mode)

        with open("out.log", "w") as f:
            f.write(result.output)

        if result.exception and not standalone_mode:
            raise result.exception

        return result

    def call_raises(self, *args, **kwargs):
        return self.call(*args, standalone_mode=True, **kwargs)

    def clean(self):
        for toclean in self.toclean:
            shutil.rmtree(toclean, ignore_errors=True)


@pytest.fixture()
def invoker(saved_cwd):
    # Snakemake 4.7 waits 10 seconds during shutdown of cluster submitted
    # worklows -- unless this is set:
    os.environ['CIRCLECI'] = "true"

    invoker = Invoker()
    yield invoker
    invoker.clean()


@pytest.fixture()
def invoker_nodir():
    # Snakemake 4.7 waits 10 seconds during shutdown of cluster submitted
    # worklows -- unless this is set:
    os.environ['CIRCLECI'] = "true"

    invoker = Invoker()
    yield invoker
    invoker.clean()


@pytest.fixture(name="envvar")
def envvar_():
    to_restore = {}

    def envvar(var, value):
        if var in os.environ:
            to_restore[var] = os.environ[var]
        else:
            to_restore[var] = None
        os.environ[var] = value

    yield envvar

    for var, value in to_restore.items():
        if value is not None:
            os.environ[var] = value
        else:
            del os.environ[var]


class MockFileDownloader(object):
    def __init__(self, block_size=None, timeout=None, parallel=None,
                 loglevel=None, alturls=None):
        pass

    def get(self, urls, dest, md5s=None):
        return True


@pytest.fixture()
def mock_downloader():
    import ymp.download
    orig = ymp.download.FileDownloader
    ymp.download.FileDownloader = MockFileDownloader
    yield MockFileDownloader
    ymp.download.FileDownloader = orig



"""
This module manages the conda environments.
"""

import io
import logging
import os
import os.path as op
import shutil
import subprocess
from typing import Optional, Union

from ruamel.yaml import YAML

import snakemake
import snakemake.conda
from snakemake.rules import Rule

import ymp
from ymp.common import AttrDict, ensure_list
from ymp.exceptions import YmpRuleError, YmpWorkflowError
from ymp.snakemake import BaseExpander, WorkflowObject


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Env(WorkflowObject, snakemake.conda.Env):
    """Represents YMP conda environment

    Snakemake expects the conda environments in a per-workflow
    directory configured by ``conda_prefix``. YMP sets this value by
    default to ``~/.ymp/conda``, which has a greater chance of being
    on the same file system as the conda cache, allowing for hard
    linking of environment files.

    Within the folder ``conda_prefix``, each environment is created in
    a folder named by the hash of the environment definition file's
    contents and the ``conda_prefix`` path. This class inherits from
    ``snakemake.conda.Env`` to ensure that the hash we use is
    identical to the one Snakemake will use during workflow execution.

    The class provides additional features for updating environments,
    creating environments dynamically and executing commands within
    those environments.

    Note:
      This is not called from within the execution. Snakemake instanciates
      its own Env object purely based on the filename.

    """

    @staticmethod
    def get_installed_env_hashes():
        cfg = ymp.get_config()
        return [
            dentry.name
            for dentry in os.scandir(cfg.absdir.conda_prefix)
            if dentry.is_dir()
        ]

    def __new__(cls, *args, **kwargs):
        if args and "name" not in kwargs:
            for env in cls.get_registry().values():
                if env.file == args[0]:
                    return env
        return super().__new__(cls)

    def __init__(self, env_file: Optional[str] = None,
                 dag: Optional[object] = None,
                 singularity_img=None,
                 name: Optional[str] = None,
                 packages: Optional[Union[list, str]] = None,
                 base: str = "none",
                 channels: Optional[Union[list, str]] = None,
                 rule: Optional[Rule] = None) -> None:
        """Creates an inline defined conda environment

        Args:
          name: Name of conda environment (and basename of file)
          packages: package(s) to be installed into environment. Version
            constraints can be specified in each package string separated from
            the package name by whitespace. E.g. ``"blast =2.6*"``
          channels: channel(s) to be selected for the environment
          base: Select a set of default channels and packages to be added to
            the newly created environment. Sets are defined in conda.defaults
            in ``yml.yml``
        """
        if 'name' in self.__dict__:
            return
        cfg = ymp.get_config()

        pseudo_dag = AttrDict({
            'workflow': {
                'persistence': {
                    'conda_env_path': cfg.ensuredir.conda_prefix,
                    'conda_env_archive_path': cfg.ensuredir.conda_archive_prefix
                }
            }
        })

        # must have either name or env_file:
        if (name and env_file) or not (name or env_file):
            raise YmpRuleError(
                self,
                "Env must have exactly one of `name` and `file`")

        if name:
            self.name = name
        else:
            self.name, _ = op.splitext(op.basename(env_file))

        if env_file:
            self.dynamic = False
            self.filename = env_file
            self.lineno = 1
        else:
            self.dynamic = True

            env_file = op.join(cfg.ensuredir.dynamic_envs, f"{name}.yml")
            defaults = {
                'name': self.name,
                'dependencies': list(ensure_list(packages) +
                                     cfg.conda.defaults[base].dependencies),
                'channels': list(ensure_list(channels) +
                                 cfg.conda.defaults[base].channels)
            }
            yaml = YAML(typ='rt')
            yaml.default_flow_style = False
            buf = io.StringIO()
            yaml.dump(defaults, buf)
            contents = buf.getvalue()

            disk_contents = ""
            if op.exists(env_file):
                with open(env_file, "r") as inf:
                    disk_contents = inf.read()
            if contents != disk_contents:
                with open(env_file, "w") as out:
                    out.write(contents)

        super().__init__(env_file, pseudo_dag, singularity_img)
        self.register()

    def set_prefix(self, prefix):
        self._env_dir = op.abspath(prefix)

    def create(self, dryrun=False):
        """Ensure the conda environment has been created

        Inherits from snakemake.conda.Env.create

        Behavior of super class
            The environment is installed in a folder in ``conda_prefix``
            named according to a hash of the ``environment.yaml`` defining
            the environment and the value of ``conda-prefix``
            (``Env.hash``). The latter is included as installed
            environments cannot be moved.

            - If this folder (``Env.path``) exists, nothing is done.

            - If a folder named according to the hash of just the contents
              of ``environment.yaml`` exists, the environment is created by
              unpacking the tar balls in that folder.

        Handling pre-computed environment specs
            In addition to freezing environments by maintaining a copy of
            the package binaries, we allow maintaining a copy of the
            package binary URLs, from which the archive folder is populated
            on demand.

        If a file ``{Env.name}.txt`` exists in ``conda.spec``
        FIXME
        """
        # Skip if environment already exists
        if op.exists(self.path):
            log.info("Environment '%s' already exists", self.name)
            return self.path

        log.warning("Creating environment '%s'", self.name)
        log.debug("Target dir is '%s'", self.path)

        # check if we have archive
        files = self._get_env_from_archive()

        # no archive? try spec
        if not files:
            urls, files, md5s = self._get_env_from_spec()
            if files:
                if dryrun:
                    log.info("Would download %i files", len(urls))
                else:
                    self._download_files(urls, md5s)
                    packages = op.join(self.archive_file, "packages.txt")
                    with open(packages, "w") as f:
                        f.write("\n".join(files) + "\n")

        # still nothing?
        if not files:
            log.warning("Neither spec file nor package archive found for '%s',"
                        " falling back to native resolver", self.name)
        else:
            if dryrun:
                log.info("Would install %i packages", len(files))
            else:
                self._install_env(files)

        res = super().create(dryrun)
        log.info("Created env %s", self.name)
        return res

    def _get_env_from_archive(self):
        """Parses Snakemake environment archive spec

        Returns:
          files: List of package archive files
        """
        packages_txt = op.join(self.archive_file, "packages.txt")
        if not op.exists(packages_txt):
            return []
        with open(packages_txt) as f:
            packages = [package.strip() for package in f]
        missing_packages = [package for package in packages
                            if not op.exists(op.join(self.archive_file,
                                                     package))]
        if missing_packages:
            log.warning(
                "Ignoring incomplete package archive for environment %s",
                self.name)
            log.debug(
                "Missing packages: %s", missing_packages)
            return []
        return packages

    def _get_env_from_spec(self):
        """Parses conda spec file

        Conda spec files contain a list of URLs pointing to the packages
        comprising the environment. Each URL may have an md5 sum appended
        as "anchor" using "#". Comments are placed at the top in lines
        beginnig with "#" and a single line "@EXPLICIT" indicates the type
        of the file.

        Returns:
          urls: list of URLs
          files: list of file names
          md5s: list of md5 sums
        """
        cfg = ymp.get_config()
        if not cfg.conda.env_specs:
            return [], [], []

        spec_file = op.join(ymp._env_dir, cfg.conda.env_specs,
                            cfg.platform, self.name + ".txt")
        if not op.exists(spec_file):
            return [], [], []

        log.debug("Using env spec '%s'", spec_file)

        with open(spec_file) as sf:
            urls = [line for line in sf
                    if line and line[0] != "@" and line[0] != "#"]
        md5s = [url.split("#")[1] for url in urls]
        files = [url.split("#")[0].split("/")[-1] for url in urls]

        return urls, files, md5s

    def _download_files(self, urls, md5s):
        from ymp.download import FileDownloader
        if not op.exists(self.archive_file):
            os.makedirs(self.archive_file)
        cfg = ymp.get_config()
        fd = FileDownloader(alturls=cfg.conda.alturls)
        if not fd.get(urls, self.archive_file, md5s):
            # remove partially download archive folder?
            # shutil.rmtree(self.archive_file, ignore_errors=True)
            raise YmpWorkflowError(
                f"Unable to create environment {self.name}, "
                f"because downloads failed. See log for details.")

    def _install_env(self, files):
        # Re-implemented part of Snakemake because we don't want "--copy"
        # forcing the installed environments to be file copies rather
        # than hardlinks where possible.
        log.info("Installing environment '%s' from %i package files",
                 self.name, len(files))
        log.debug("Path: %s", self.archive_file)
        log.debug("Files: %s", files)
        log.info("Calling conda...")
        install_files = [op.join(self.archive_file, fn) for fn in files]

        sp = subprocess.run(["conda", "create", "--prefix",
                             self.path] + install_files)
        if sp.returncode != 0:
            # try make sure we don't leave partially installed env around
            shutil.rmtree(self.path, ignore_errors=True)
            raise YmpWorkflowError(
                f"Unable to create environment {self.name}, "
                f"because conda create failed"
            )
        log.info("Conda complete")

    @property
    def installed(self):
        return op.isdir(self.path)

    def update(self):
        "Update conda environment"
        self.create()  # call create to make sure environment exists
        log.warning("Updating environment '%s'", self.name)
        return subprocess.run([
            "conda", "env", "update",
            "--prune",
            "-p", self.path,
            "-f", self.file,
            "-v"
        ]).returncode

    def run(self, command):
        """Execute command in environment

        Returns exit code of command run.
        """
        cfg = ymp.get_config()
        command = "source activate {}; {}".format(self.path, " ".join(command))
        log.debug("Running: %s", command)
        return subprocess.run(command,
                              shell=True, executable=cfg.shell).returncode

    def export(self, stream, typ='yml'):
        """Freeze environment"""
        log.warning("Exporting environment '%s'", self.name)
        if typ == 'yml':
            res = subprocess.run([
                "conda", "env", "export",
                "-p", self.path,
            ], stdout=subprocess.PIPE)

            yaml = YAML(typ='rt')
            yaml.default_flow_style = False
            env = yaml.load(res.stdout)
            env['name'] = self.name
            if 'prefix' in env:
                del env['prefix']
            yaml.dump(env, stream)
        elif typ == 'txt':
            res = subprocess.run([
                "conda", "list", "--explicit", "--md5",
                "-p", self.path,
            ], stdout=stream)
        return res.returncode

    def __lt__(self, other):
        "Comparator for sorting"
        return self.name < other.name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__dict__!r})"

    def __eq__(self, other):
        if isinstance(other, Env):
            return self.hash == other.hash


# Patch Snakemake's Env class with our own
snakemake.conda.Env = Env


class CondaPathExpander(BaseExpander):
    """Applies search path for conda environment specifications

    File names supplied via ``rule: conda: "some.yml"`` are replaced with
    absolute paths if they are found in any searched directory.
    Each ``search_paths`` entry is appended to the directory
    containing the top level Snakefile and the directory checked for
    the filename. Thereafter, the stack of including Snakefiles is traversed
    backwards. If no file is found, the original name is returned.
    """
    def __init__(self, config, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._search_paths = config.conda.env_path
        self._envs = None

    def expands_field(self, field):
        return field == 'conda_env'

    def format(self, conda_env, *args, **kwargs):
        if conda_env[0] == "/" and op.exists(conda_env):
            return conda_env
        if not self._envs:
            self._envs = Env.get_registry()
        if conda_env in self._envs:
            return self._envs[conda_env].file

        for snakefile in reversed(self.workflow.included_stack):
            basepath = op.dirname(snakefile)
            for _, relpath in sorted(self._search_paths.items()):
                searchpath = op.join(basepath, relpath)
                abspath = op.abspath(op.join(searchpath, conda_env))
                for ext in "", ".yml", ".yaml":
                    env_file = abspath+ext
                    if op.exists(env_file):
                        Env(env_file, rule=self.current_rule)
                        return env_file
        return conda_env



"""This module contains a Sphinx_ extension for documenting YMP stages and
Snakemake_ rules.

The `SnakemakeDomain` (name **sm**) provides the following directives:

.. rst:directive:: .. sm:rule:: name

   Describes a `Snakemake rule <snakefiles-rules>`

.. rst:directive:: .. sm:stage:: name

   Describes a `YMP Stage <Stage>`

Both directives accept an optional ``source`` parameter. If given, a
link to the source code of the stage or rule definition will be added.
The format of the string passed is ``filename:line``. Referenced
Snakefiles will be highlighted with pygments and added to the
documentation when building HTML.

The extension also provides an autodoc-like directive:

.. rst:directive:: .. autosnake:: filename

   Generates documentation from Snakefile ``filename``.

.. _Sphinx: http://sphinx-doc.org
.. _Snakemake: http://snakemake.readthedocs.io

"""

import os
from textwrap import dedent, indent
from typing import List, Optional

from docutils import nodes
from docutils.parsers import rst
from docutils.statemachine import StringList

from snakemake.rules import Rule

from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.environment.collectors import EnvironmentCollector
from sphinx.roles import XRefRole
from sphinx.util import logging, ws_re
from sphinx.util.nodes import make_refnode

import ymp
from ymp.snakemake import ExpandableWorkflow
from ymp.snakemakelexer import SnakemakeLexer
from ymp.stage import Stage

try:
    logger = logging.getLogger(__name__)
except AttributeError:
    # Fall back to normal logging
    import logging as _logging
    logger = _logging.getLogger(__name__)

#: str: Path in which YMP package is located
BASEPATH = os.path.dirname(os.path.dirname(ymp.__file__))


def relpath(path: str) -> str:
    """Make absolute path relative to BASEPATH

    Args:
      path: absolute path

    Returns:
      path relative to BASEPATH
    """
    return os.path.relpath(path, BASEPATH)


class YmpObjectDescription(ObjectDescription):
    """
    Base class for RSt directives in SnakemakeDomain

    Since this inherhits from Sphinx' ObjectDescription, content
    generated by the directive will always be inside an addnodes.desc.

    Args:
      source: Specify source position as ``file:line`` to create link
    """
    typename = "[object name]"

    option_spec = {
        # source link (<filename>:<lineno>)
        'source': rst.directives.unchanged
    }

    def handle_signature(self, sig: str, signode: addnodes.desc) -> str:
        """
        Parse rule signature *sig* into RST nodes and append them
        to *signode*.

        The retun value identifies the object and is passed to
        :meth:`add_target_and_index()` unchanged

        Args:
          sig: Signature string (i.e. string passed after directive)
          signode: Node created for object signature
        Returns:
          Normalized signature (white space removed)
        """
        signode += addnodes.desc_annotation(self.typename, self.typename+" ")
        signode += addnodes.desc_name(sig, sig)

        if 'source' in self.options:
            self.add_source_link(signode)

        sigid = ws_re.sub('', sig)
        return sigid

    def add_source_link(self, signode: addnodes.desc) -> None:
        """
        Add link to source code to *signode*
        """
        filename, lineno = self.options['source'].split(':')
        if not hasattr(self.env, '_snakefiles'):
            self.env._snakefiles = set()
        self.env._snakefiles.add(filename)

        onlynode = addnodes.only(expr='html')  # show only in html
        onlynode += nodes.reference(
            '',
            refuri='_snakefiles/{}.html#line-{}'.format(filename, lineno)
        )
        onlynode[0] += nodes.inline('', '[source]',
                                    classes=['viewcode-link'])
        signode += onlynode

    def add_target_and_index(self, name: str, sig: str,
                             signode: addnodes.desc) -> None:
        """
        Add cross-reference IDs and entries to ``self.indexnode``
        """
        targetname = "-".join((self.objtype, name))
        if targetname not in self.state.document.ids:
            signode['names'].append(targetname)
            signode['ids'].append(targetname)
            signode['first'] = (not self.names)
            self.state.document.note_explicit_target(signode)

            objects = self.env.domaindata[self.domain]['objects']
            key = (self.objtype, name)
            if key in objects:
                self.env.warn(self.env.docname,
                              'duplicate description of {} {}, '
                              'other instance in {}:{}'
                              ''.format(self.objtype, name,
                                        self.env.doc2path(objects[key][0]),
                                        self.lineno))
            objects[key] = (self.env.docname, targetname)

        # register rule in index
        indextext = self.get_index_text(self.objtype, name)
        if indextext:
            self.indexnode['entries'].append((
                'single',
                indextext,
                targetname,
                '',
                None))

    def get_index_text(self, typename: str, name: str) -> str:
        """Formats object for entry into index"""
        return "{} ({})".format(name, typename)


class SnakemakeRule(YmpObjectDescription):
    """
    Directive ``sm:rule::`` describing a Snakemake rule
    """
    typename = "rule"


class YmpStage(YmpObjectDescription):
    """
    Directive ``sm:stage::`` describing an YMP stage
    """
    typename = "stage"


class SnakemakeDomain(Domain):
    """Snakemake language domain"""
    name = "sm"
    label = "Snakemake"

    object_types = {
        # ObjType(name, *roles, **attrs)
        'rule': ObjType('rule', 'rule'),
        'stage': ObjType('stage', 'stage'),
    }
    directives = {
        'rule': SnakemakeRule,
        'stage': YmpStage,
    }
    roles = {
        'rule': XRefRole(),
        'stage': XRefRole(),
    }
    initial_data = {
        'objects': {},  #: (type, name) -> docname, labelid
    }

    data_version = 0

    def clear_doc(self, docname: str):
        """Delete objects derived from file ``docname``"""
        if 'objects' in self.data:
            toremove = [
                key
                for (key, (docname_, _)) in self.data['objects'].items()
                if docname_ == docname
            ]
            for key in toremove:
                del self.data['objects'][key]

    def resolve_xref(self, env: BuildEnvironment, fromdocname: str,
                     builder, typ, target, node, contnode):
        objects = self.data['objects']
        objtypes = self.objtypes_for_role(typ)
        for objtype in objtypes:
            if (objtype, target) in objects:
                return make_refnode(builder, fromdocname,
                                    objects[objtype, target][0],
                                    objects[objtype, target][1],
                                    contnode, target + ' ' + objtype)

    def get_objects(self):
        for (typ, name), (docname, ref) in self.data['objects'].items():
            # name, dispname, type, docname, anchor, searchprio
            yield name, name, typ, docname, ref, 1


class AutoSnakefileDirective(rst.Directive):
    """Implements RSt directive ``.. autosnake:: filename``

    The directive extracts docstrings from rules in snakefile and
    auto-generates documentation.
    """

    #: bool: This rule does not accept content
    has_content = False

    #: int: This rule needs one argument (the filename)
    required_arguments = 1

    #: str: Template for generated Rule RSt
    tpl_rule = ".. sm:rule:: {name}"

    #: str: Template for generated Stage RSt
    tpl_stage = ".. sm:stage:: {name}"

    #: str: Template option source
    tpl_source = "   :source: {filename}:{lineno}"

    def run(self):
        """Entry point"""
        snakefile = self.arguments[0]

        #: BuildEnvironment: Sphinx build environment
        self.env: BuildEnvironment = self.state.document.settings.env

        #: ExpandableWorkflow: Ymp Workflow object
        self.workflow = self.load_workflow(snakefile)

        return self._generate_nodes()

    def load_workflow(self, file_path: str) -> ExpandableWorkflow:
        """Load the Snakefile"""
        workflow = ExpandableWorkflow(snakefile=file_path)
        workflow.include(file_path)
        return workflow

    def parse_doc(self, doc: str, source: str, idt: int=0) -> StringList:
        """Convert doc string to StringList

        Args:
          doc: Documentation text
          source: Source filename
          idt: Result indentation in characters (default 0)

        Returns:
          StringList of re-indented documentation wrapped in newlines
        """
        doc = dedent(doc or "").strip("\n")
        doc = indent(doc, " " * idt)
        doclines = [''] + doc.splitlines() + ['']
        return StringList(doclines, source)

    def parse_rule(self, rule: Rule, idt: int=0) -> StringList:
        """Convert Rule to StringList

        Args:
          rule: Rule object
          idt: Result indentation in characters (default 0)

        Retuns:
          StringList containing formatted Rule documentation
        """
        head = self.tpl_rule.format(name=rule.name)
        if rule.lineno:
            head += "\n"
            head += self.tpl_source.format(
                filename=relpath(rule.snakefile),
                lineno=self.workflow.linemaps[rule.snakefile][rule.lineno],
            )
        head = indent(head, " " * idt)
        headlines = head.splitlines()
        doc = self.parse_doc(rule.docstring, rule.snakefile, idt+3)

        return StringList(headlines, rule.snakefile) + doc

    def parse_stage(self, stage: Stage, idt: int=0) -> StringList:
        head = self.tpl_stage.format(name=stage.name)
        if stage.lineno:
            head += "\n"
            head += self.tpl_source.format(
                filename=relpath(stage.filename),
                lineno=self.workflow.linemaps[stage.filename][stage.lineno],
            )
        head = indent(head, " " * idt)
        headlines = head.splitlines()
        doc = self.parse_doc(stage.docstring, stage.filename, idt+3)

        res = StringList(headlines, stage.filename) + doc
        for rule in sorted(stage.rules, key=lambda s: s.name):
            res.extend(self.parse_rule(rule, idt+3))
        return res

    def _generate_nodes(self) -> List[nodes.Node]:
        """Generate Sphinx nodes from parsed snakefile"""
        node = nodes.paragraph('')
        result = StringList()

        # generate stages
        stages = Stage.get_registry().values()
        stages = list(set(stages))
        stages = sorted(stages, key=lambda x: x.name)
        for stage in stages:
            result.extend(self.parse_stage(stage))

        # generate nodes for rules not registered with stages
        rules = self.workflow.rules
        for rule in rules:
            if not getattr(rule, "ymp_stage", False):
                result.extend(self.parse_rule(rule))

        self.state.nested_parse(result, 0, node)
        return [node]


def collect_pages(app: Sphinx):
    """Add Snakefiles to documentation (in HTML mode)
    """
    if not hasattr(app.env, '_snakefiles'):
        return

    highlight_block = app.builder.highlighter.highlight_block

    for snakefile in app.env._snakefiles:
        try:
            with open(os.path.join(BASEPATH, snakefile), 'r') as f:
                code = f.read()
        except IOError:
            logger.error("failed to open {}".format(snakefile))
            continue
        highlighted = highlight_block(code, 'snakemake', lineanchors="line")
        context = {
            'title': snakefile,
            'body': '<h1>Snakefile "{}"</h1>'.format(snakefile) +
            highlighted
        }
        yield (os.path.join('_snakefiles', snakefile), context, 'page.html')

    html = ['\n']
    context = {
        'title': ('Overview: Snakemake rule files'),
        'body': '<h1>All Snakemake rule files</h1>' +
        ''.join(html)
    }
    yield ('_snakefiles/index', context, 'page.html')


class DomainTocTreeCollector(EnvironmentCollector):
    """Add Sphinx Domain entries to the TOC"""

    # override
    def clear_doc(self, app: Sphinx,
                  env: BuildEnvironment, docname: str) -> None:
        """Clear data from environment

        If we have cached data in environment for document ``docname``,
        we should clear it here.

        """

    # override
    def merge_other(self, app: Sphinx, env: BuildEnvironment,
                    docnames: List[str], other: BuildEnvironment) -> None:
        """Merge with results from parallel processes

        Called if Sphinx is processing documents in parallel. We
        should merge this from ``other`` into ``env`` for all ``docnames``.
        """

    # override
    def process_doc(self, app: Sphinx, doctree: nodes.Node) -> None:
        """Process ``doctree``

        This is called by ``read-doctree``, so after the doctree has been
        loaded. The signal is processed in registered first order,
        so we are called after built-in extensions, such as the
        :mod:`sphinx.environment.collectors.toctree` extension building
        the TOC.
        """

        # FIXME: handle duplicate entries
        for node in self.select_doc_nodes(doctree):
            tocnode = self.select_toc_location(app, node)
            heading = self.make_heading(node)
            if not tocnode:
                continue

            self.toc_insert(app.env.docname, tocnode, node, heading)

    def select_doc_nodes(self, doctree: nodes.Node) -> List[nodes.Node]:
        """Select the nodes for which entries in the TOC are desired

        This is a separate method so that it might be overriden by
        subclasses wanting to add other types of nodes to the TOC.
        """
        return doctree.traverse(addnodes.desc)

    def select_toc_location(self, app: Sphinx,
                            node: nodes.Node) -> nodes.Node:
        """Select location in TOC where ``node`` should be referenced
        """
        while node is not None:
            tocnode = self.locate_in_toc(app, node)
            if tocnode:
                return tocnode

            node = node.parent
        return app.env.tocs[app.env.docname][0]

    def locate_in_toc(self, app: Sphinx,
                      node: nodes.Node) -> Optional[nodes.Node]:
        toc = app.env.tocs[app.env.docname]
        ref = self.get_ref(node)
        for node in toc.traverse(nodes.reference):
            node_ref = node.get('anchorname')
            if not node_ref or node_ref[0] != "#":
                continue
            if node_ref[1:] == ref:
                return node.parent.parent

    def get_ref(self, node: nodes.Node) -> Optional[nodes.Node]:
        while node is not None:
            if not node.get('ids'):
                # In Sphinx domain descriptions, the ID is in the
                # first child node, the desc_signature.
                # We can take that, but to make the JS handling
                # the TOC sidebar work properly, we need it to
                # be in the hierarchy. So create an ID for the
                # desc node.
                if node[0].get('ids'):
                    node['ids'] = [node[0].get('ids')[0] + '-tocentry']
            if node.get('ids'):
                return node['ids'][0]
            node = node.parent

    def make_heading(self, node: nodes.Node) -> List[nodes.Node]:
        names = node[0].traverse(addnodes.desc_name)
        return names

    def toc_insert(self, docname: str, tocnode: nodes.Node, node: nodes.Node,
                   heading: List[nodes.Node]) -> None:
        for child in tocnode.children:
            if isinstance(child, nodes.bullet_list):
                blist = child
                break
        else:
            blist = nodes.bullet_list('')
            tocnode += blist

        reference = nodes.reference(
            '', '', internal=True, refuri=docname,
            anchorname="#" + self.get_ref(node), *heading)
        para = addnodes.compact_paragraph('', '', reference)
        item = nodes.list_item('', para)
        # FIXME: find correct location
        blist.append(item)


def setup(app: Sphinx):
    """Register the extension with Sphinx"""
    app.add_lexer('snakemake', SnakemakeLexer())
    app.add_domain(SnakemakeDomain)
    app.add_directive('autosnake', AutoSnakefileDirective)
    app.add_env_collector(DomainTocTreeCollector)
    app.connect('html-collect-pages', collect_pages)



import re
import string
from itertools import product

import pytest

import ymp.string

MERGED = 0
PRODUCT = 1


class FormatterTest(object):
    pattern = "{A} and {B}"
    fields = {'A': [1, 2], 'B': [3, 4]}
    result = '[1, 2] and [3, 4]'
    formatter = string.Formatter()

    def format(self, pattern, *args, **kwargs):
        return self.formatter.format(pattern, *args, **kwargs)

    def test_formatting(self):
        assert self.format(self.pattern, **self.fields) == self.result


class OverrideJoinFormatterTest(FormatterTest):
    formatter = ymp.string.OverrideJoinFormatter()


class ProductFormatterTest(FormatterTest):
    formatter = ymp.string.ProductFormatter()
    result = ['1 and 3', '1 and 4', '2 and 3', '2 and 4']


class RegexFormatterTest(FormatterTest):
    regex = re.compile(
        r"""
        \{
            (?=(
                (?P<name>[^{}]+)
            ))\1
        \}
        """, re.VERBOSE)
    formatter = ymp.string.RegexFormatter(regex)


class PartialFormatterTest(FormatterTest):
    formatter = ymp.string.PartialFormatter()

    def test_partial_formatting(self):
        pat = self.pattern
        for key in self.fields:
            pat = self.format(pat, **{key: self.fields[key]})
        assert pat == self.result


@pytest.fixture(params=[
    "{A} and {B}"
])
def pattern(request):
    return request.param


@pytest.fixture(params=[
    {'A': [1, 2], 'B': [3, 4]},
    {'A': ['1', '2'], 'B': ['3', '4']},
    {'A': 1, 'B': [3, 4]},
    {'A': '12', 'B': [3, 4]}
], ids=[
    'numbers',
    'strings',
    'int',
    'string'
])
def values(request):
    return request.param


@pytest.fixture(params=[product, None], ids=['Product', 'Flatten'])
def combine(request, pattern, values):
    if request.param is None:
        result = pattern.format(**values)
    else:
        result = [
            pattern.format(**dict(zip(values, x)))
            for x in request.param(*(value
                                     if isinstance(value, list)
                                     else (value,)
                                     for value in values.values()))
        ]

    return {
        'product': 1 if request.param is not None else None,
        'result': result,
        'pattern': pattern,
        'values': values
    }


@pytest.fixture(params=[
    (None, "{", "}"),
    (r"\{(?=((?P<name>[^{}]+)))\1\}", "{", "}"),
    (r"\{:(?=(\s*(?P<name>(?:.(?!\s*\:\}))*.)\s*))\1:\}", "{:", ":}"),
], ids=[
    'NoRegex',
    'Braces',
    'BraceColon'
])
def combine_regex(request, combine):
    combine['regex'] = request.param[0]
    combine['pattern'] = (combine['pattern']
                          .replace("{", request.param[1])
                          .replace("}", request.param[2])
                          )
    return combine


def test_make_formatter(combine_regex):
    do_product = combine_regex['product']
    do_regex = combine_regex['regex']
    values = combine_regex['values']
    pattern = combine_regex['pattern']
    expected_result = combine_regex['result']

    formatter = ymp.string.make_formatter(
        product=do_product,
        regex=do_regex,
        partial=None)
    result = formatter.format(pattern, **values)
    assert result == expected_result



import io
import logging
import os
from collections.abc import (
    ItemsView, KeysView, Mapping, MappingView, Sequence, ValuesView
)

import ruamel.yaml
from ruamel.yaml import RoundTripRepresenter, YAML
from ruamel.yaml.comments import CommentedMap


log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class MixedTypeError(Exception):
    """Mixed types in proxy collection"""


class LayeredConfError(Exception):
    """Error in LayeredConf"""


class LayeredConfWriteError(LayeredConfError):
    """Can't write"""


class LayeredConfAccessError(LayeredConfError, KeyError, IndexError):
    """Can't access"""


class AttrItemAccessMixin(object):
    """Mixin class mapping dot to bracket access

    Added to classes implementing __getitem__, __setitem__ and
    __delitem__, this mixin will allow acessing items using dot
    notation. I.e. "object.xyz" is translated to "object[xyz]".
    """
    def __getattr__(self, key):
        try:
            if key[0] == "_":
                return self.__getattribute__(key)
            else:
                return self[key]
        except (IndexError, KeyError) as e:
            raise AttributeError(e)

    def __setattr__(self, key, value):
        try:
            if key[0] == "_":
                object.__setattr__(self, key, value)
            else:
                self[key] = value
        except (IndexError, KeyError) as e:
            raise AttributeError(e)

    def __delattr__(self, key):
        raise NotImplementedError()


class MultiProxy(object):
    """Base class for layered container structure"""
    def __init__(self, maps, parent=None, key=None):
        self._maps = list(maps)
        self._parent = parent
        self._key = key

    def make_map_proxy(self, key, items):
        return MultiMapProxy(items, parent=self, key=key)

    def make_seq_proxy(self, key, items):
        return MultiSeqProxy(items, parent=self, key=key)

    def get_files(self):
        return [fn for fn, layer in self._maps]

    def to_yaml(self, show_source=False):
        buf = io.StringIO()
        if show_source:
            for fn, layer in self._maps:
                buf.write(f"--- # from '{fn}' # ---\n")
                rt_yaml.dump(layer, buf)
        else:
            rt_yaml.dump(self, buf)
        return buf.getvalue()

    def __str__(self):
        return self.to_yaml()

    def __repr__(self):
        return f"{self.__class__.__name__}({self._maps!r})"

    def add_layer(self, name, container):
        self._maps.insert(0, ((name, container)))

    def remove_layer(self, name):
        map_name = self._maps[0][0]
        if map_name == name:
            self._maps.pop(0)
        else:
            raise LayeredConfError(f"in remove_layer: {map_name} != {name}")


class MultiMapProxyMappingView(MappingView):
    """MappingView for MultiMapProxy"""
    def __init__(self, mapping):
        self._mapping = mapping

    def __len__(self):
        return len(self._mapping)

    def __repr__(self):
        return '{0.__class__.__name__}({0._mapping!r})'.format(self)

    def __radd__(self, other):
        if isinstance(other, Sequence):
            return other + list(self)
        raise TypeError()


class MultiMapProxyItemsView(MultiMapProxyMappingView, ItemsView):
    """ItemsView for MultiMapProxy"""
    def __contains__(self, key):
        return (isinstance(key, tuple)
                and len(key) == 2
                and key[0] in self._mapping
                and self._mapping[key[0]] == key[1])

    def __iter__(self):
        for key in self._mapping:
            yield key, self._mapping[key]


class MultiMapProxyKeysView(MultiMapProxyMappingView, KeysView):
    """KeysView for MultiMapProxy"""
    def __contains__(self, key):
        return key in self._mapping

    def __iter__(self):
        yield from iter(self._mapping)


class MultiMapProxyValuesView(MultiMapProxyMappingView, ValuesView):
    """ValuesView for MultiMapProxy"""
    def __contains__(self, key):
        return any(self._mapping[k] == key for k in self._mapping)

    def __iter__(self):
        for key in self._mapping:
            yield self._mapping[key]


class MultiMapProxy(Mapping, MultiProxy, AttrItemAccessMixin):
    """Mapping Proxy for layered containers"""
    def __contains__(self, key):
        return any(key in m for _, m in self._maps)

    def __len__(self):
        return len(set(k for _, m in self._maps for k in m))

    def __getitem__(self, key):
        items = [(fn, m[key]) for fn, m in self._maps if key in m]
        if not items:
            raise KeyError(f"key '{key}' not found in any map")
        typs = set(type(m[1]) for m in items if m[1])
        if len(typs) > 1:
            raise MixedTypeError(
                f"while trying to obtain '{key}' from {items!r},"
                f"types differ: {typs}"
            )
        if isinstance(items[0][1], Mapping):
            return self.make_map_proxy(key, items)
        if isinstance(items[0][1], str):
            return items[0][1]
        if isinstance(items[0][1], Sequence):
            return self.make_seq_proxy(key, items)
        return items[0][1]

    def __setitem__(self, key, value):
        # we want to set to the top layer, so get that first
        mp = self
        keys = []
        while mp._parent:
            keys.append(mp._key)
            mp = mp._parent
        # now walk back down
        for k in reversed(keys):
            if k not in mp._maps[0][1]:
                mp._maps[0][1][k] = CommentedMap()
            mp = mp[k]
        # and set the value, potentially on a different object than self
        mp._maps[0][1][key] = value

    def __delitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        for key in set(k for _, m in self._maps for k in m):
            yield key

    def get(self, value, default=None):
        try:
            return self[value]
        except KeyError:
            return default

    def items(self):
        return MultiMapProxyItemsView(self)

    def keys(self):
        return MultiMapProxyKeysView(self)

    def values(self):
        return MultiMapProxyValuesView(self)


class MultiSeqProxy(Sequence, MultiProxy, AttrItemAccessMixin):
    """Sequence Proxy for layered containers"""
    def __contains__(self, value):
        return any(value in m for _, m in self._maps)

    def __iter__(self):
        for _, m in self._maps:
            for item in m:
                yield item

    def __len__(self):
        return sum(len(m) for _, m in self._maps)

    def __repr__(self):
        return f"{self.__class__.__name__}({self._maps})"

    def __str__(self):
        return "+".join(f"{m}" for _, m in self._maps)

    def __getitem__(self, index):
        if isinstance(index, slice):
            raise NotImplementedError()
        if isinstance(index, str):
            try:
                index = int(index)
            except ValueError:
                raise KeyError()
        for _, m in self._maps:
            if index >= len(m):
                index -= len(m)
            else:
                return m[index]

    def __radd__(self, other):
        return self.__add__(other)

    def __add__(self, other):
        return other + list(self)

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __delitem__(self, key):
        raise NotImplementedError()

    def __missing__(self, key):
        raise NotImplementedError()

    def __iadd__(self, item):  # broken?!
        self.extend(item)

    def extend(self, item):
        m = self._maps[0][1]
        m.extend(item)


class LayeredConfProxy(MultiMapProxy):
    """Layered configuration"""

    def __str__(self):
        try:
            return self.to_yaml()
        except ruamel.yaml.serializer.SerializerError:
            return self.__repr__()

    def __enter__(self):
        self.add_layer("dynamic", {})
        return self

    def __exit__(self, *args):
        self.remove_layer("dynamic")

    def save(self, outstream=None, layer=0):
        outfile = None
        if outstream:
            rt_yaml.dump(self._maps[layer][1], outstream)
        else:
            outfile = self._maps[layer][0]
            with open(outfile+".tmp", "w") as outstream:
                rt_yaml.dump(self._maps[layer][1], outstream)
            os.rename(outfile+".tmp", outfile)


RoundTripRepresenter.add_representer(LayeredConfProxy,
                                     RoundTripRepresenter.represent_dict)
RoundTripRepresenter.add_representer(MultiMapProxy,
                                     RoundTripRepresenter.represent_dict)
RoundTripRepresenter.add_representer(MultiSeqProxy,
                                     RoundTripRepresenter.represent_list)

rt_yaml = YAML(typ="rt")


def load(files):
    """Load configuration files

    Creates a `LayeredConfProxy` configuration object from a set of
    YAML files.
    """
    layers = []
    for fn in reversed(files):
        with open(fn, "r") as f:
            yaml = rt_yaml.load(f)
            if not isinstance(yaml, Mapping):
                raise LayeredConfError(
                    f"Malformed config file '{fn}'."
                )
            layers.append((fn, yaml))
    return LayeredConfProxy(layers)



"""
ymp.snakemakelexer
~~~~~~~~~~~~~~~~~~

"""

from pygments.lexers.python import Python3Lexer
from pygments.token import Keyword, Text, Name
from pygments.lexer import bygroups, inherit, words, include


class SnakemakeLexer(Python3Lexer):
    name = 'Snakemake'
    tokens = {
        'root': [
            (r'(rule)((?:\s|\\\s)+)', bygroups(Keyword, Text), 'rulename'),
            include('rulekeyword'),
            include('globalkeyword'),
            inherit,
        ],
        'rulename': [
            ('[a-zA-Z_]\w*', Name.Class, '#pop')
        ],
        'rulekeyword': [
            (words((
                'benchmark', 'conda', 'input', 'log', 'message', 'output',
                'params', 'priority', 'resources', 'shadow', 'shell',
                'singularity', 'threads', 'version', 'wildcard_constraints',
                'wrapper', 'run', 'script'
            ), suffix=r'(?=\s*:)'), Keyword),
        ],
        'globalkeyword': [
            (words((
                'include', 'workdir', 'configfile', 'ruleorder',
                'global_wildcard_constraints', 'subworkflow',
                'localrules', 'onsuccess', 'onerror', 'onstart'
                ), suffix=r'\b'), Keyword),
        ]
    }

# uncomment to print token stream for debug
#    def get_tokens_unprocessed(self, text):
#        import sys
#        print("\n".join(str(l) for l in super().get_tokens_unprocessed(text)), file=sys.stderr)
#        yield from super().get_tokens_unprocessed(text)



import ymp
from ymp.common import Cache


class LoadFuncs(object):
    def __init__(self):
        self.count = 0

    def reset(self):
        self.count = 0

    def make_itemfunc(self):
        def itemloadfunc(key, arg=None):
            self.count += 1
            return key
        return itemloadfunc


def test_cache_version(saved_cwd):
    fact = LoadFuncs()
    cache = Cache(saved_cwd)
    test_cache = cache.get_cache("test", itemloadfunc=fact.make_itemfunc())
    assert fact.count == 0
    assert test_cache[1] == 1
    # first run of new cache, itemfunc should be called once:
    assert fact.count == 1
    cache.close()

    fact.reset()
    cache = Cache(saved_cwd)
    test_cache = cache.get_cache("test", itemloadfunc=fact.make_itemfunc())
    assert fact.count == 0
    assert test_cache[1] == 1
    # second run of new cache, itemfunc should not have been called
    assert fact.count == 0
    cache.close()

    # alter version
    true_version = ymp.__numeric_version__
    ymp.__numeric_version__ = -1

    fact.reset()
    cache = Cache(saved_cwd)
    test_cache = cache.get_cache("test", itemloadfunc=fact.make_itemfunc())
    assert fact.count == 0
    assert test_cache[1] == 1
    # run with changed version, expecting one call
    assert fact.count == 1
    cache.close()

    # restore version
    ymp.__numeric_version__ == true_version



import os
import sys
import warnings



try:
    from ymp._version import version as __version__
except ModuleNotFoundError:
    from pkg_resources import get_distribution, DistributionNotFound
    try:
        __version__ = get_distribution(__name__).version
    except DistributionNotFound:
        from setuptools_scm import get_version
        __version__ = get_version(root="..", relative_to=__file__)


try:
    __numeric_version__ = sum(
        (100 ** n) * int(m)
        for n, m in enumerate(__version__.split(".")[2::-1]))
except:
    warnings.warn("Could not parse version {__version__}")
    __numeric_version__ = 0

# Importing pkg_resources takes rather long (~200ms), for CLI snappiness,
# we manually gather the paths for our distributed files.
_rsc_dir = __path__[0]
_rule_dir = os.path.join(_rsc_dir, "rules")
_etc_dir = os.path.join(_rsc_dir, "etc")
_snakefile = os.path.join(_rule_dir, "Snakefile")
_defaults_file = os.path.join(_etc_dir, "defaults.yml")
_env_dir = os.path.join(_rsc_dir, "conda_envs")

# import ymp.config for type hints below
# (this import will eventually access the path variables above, so moving
# it further up breaks docs build)
if 'sphinx' in sys.modules:
    import ymp.config


#: Set to 1 to show the YMP expansion process as it is applied to the next
#: Snakemake rule definition.
#:
#: >>> ymp.print_rule = 1
#: >>> rule broken:
#: >>>   ...
#:
#: >>> ymp make broken -vvv
print_rule = 0


def get_config() -> 'config.ConfigMgr':
    """Access the current YMP configuration object.

    This object might change once during normal execution: it is
    deleted before passing control to Snakemake. During unit test
    execution the object is deleted between all tests.
    """
    from ymp.config import ConfigMgr
    return ConfigMgr.instance()



import asyncio
import atexit
import hashlib
import logging
import os
import re
import threading
from typing import List, Optional, Union
from urllib.parse import urlsplit

import aiohttp

from tqdm import tqdm

from ymp.common import ensure_list

LOG = logging.getLogger(__name__)


class FileDownloader(object):
    """Manages download of a set of URLs

    Downloads happen concurrently using asyncronous network IO.

    Args:
      block_size: Byte size of chunks to download
      timeout:    Aiohttp cumulative timeout
      parallel:   Number of files to download in parallel
      loglevel:   Log level for messages send to logging
                  (Errors are send with loglevel+10)
      alturls:    List of regexps modifying URLs
      retry:      Number of times to retry download
    """
    def __init__(self, block_size: int=4096, timeout: int=300, parallel: int=4,
                 loglevel: int=logging.WARNING, alturls=None, retry: int=3):
        self._block_size = block_size
        self._timeout = timeout
        self._parallel = parallel
        self._retry = retry

        self._alturls = []
        alturls = ["///"] + (alturls or [])
        for pat in alturls:
            sep = pat[0]
            if pat.strip(sep):
                patsub = re.split(r"(?<=[^\\])"+sep, pat.strip(sep))
                if len(patsub) != 2:
                    raise ValueError("Malformed regular expression '{}'"
                                     "".format(pat))
                patsub[1] = patsub[1].replace(r"\/", "/")
            else:
                patsub = ["", ""]
            self._alturls.append(patsub)

        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            # no loop in context (i.e. running in thread)
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)

        self._sem = asyncio.Semaphore(parallel)
        self._progress = LOG.getEffectiveLevel() <= loglevel
        self._loglevel = loglevel
        self._sum_bar = None

    def log(self, msg: str, *args, modlvl: int=0, **kwargs) -> None:
        """Send message to logger

        Honors loglevel set for the FileDownloader object.

        Args:
          msg: The log message
          modlvl: Added to default logging level for object
        """
        LOG.log(self._loglevel + modlvl, msg, *args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        """Send error to logger

        Message is sent with a log level 10 higher than the default
        for this object.
        """
        self.log(msg, *args, modlvl=10, **kwargs)

    @staticmethod
    def make_bar_format(desc_width: int=20, count_width: int=0,
                        rate: bool=False, eta: bool=False,
                        have_total: bool=True) -> str:
        """Construct bar_format for tqdm

        Args:
          desc_width: minimum space allocated for description
          count_width: min space for counts
          rate: show rate to right of progress bar
          eta: show eta to right of progress bar
          have_total: whether a total exists (required to add percentage)
        """
        if have_total:
            left = '{{desc:<{dw}}} {{percentage:3.0f}}%'.format(dw=desc_width)
        else:
            # percentage not supplied by TQDM if there is no total
            left = '{{desc:<{dw}}}'.format(dw=desc_width)
        right = ' {{n_fmt:>{cw}}} / {{total_fmt:<{cw}}}'.format(cw=count_width)
        if rate:
            right += ' {{rate_fmt:>{cw}}}'.format(cw=count_width+2)
        if eta:
            right += ' ETA {remaining}'
        return left + '|{bar}|' + right

    async def _download(self, session: aiohttp.ClientSession,
                        url: str, dest: str, md5: Optional[str]=None) -> bool:
        """Asynchronously download a single file

        - If ``dest`` points to an existing directory, the file name
        is derived from the trailing path portion of the URL.

        - Will skip download for existing files with matching MD5

        Args:
          session: aiohttp session object
          url:     source url
          dest:    destination path
          md5:     optional md5 checksum to verify
        """
        if os.path.isdir(dest):
            parts = urlsplit(url)
            basename = os.path.basename(parts.path)
            destfile = os.path.join(dest, basename)
        else:
            basename = os.path.basename(dest)
            destfile = dest

        if os.path.exists(destfile) and md5 and not isinstance(md5, bool):
            if self._check_md5(basename, destfile, md5):
                return True

        tryurls = [re.sub(pat, rep, url) for pat, rep in self._alturls]
        for url in tryurls:  # try alturls
            exc = None
            for num_try in range(self._retry):  # retry after timeout
                if exc:
                    self.log("Downloading %s failed with %s. Retrying %i/%i",
                             basename, exc, num_try, self._retry-1)
                try:
                    if await self._download_one(session, basename, url,
                                                destfile, md5):
                        return True
                    break
                except TimeoutError as e:
                    exc = e
        return False

    def _check_md5(self, name, fname, md5):
        md5_new = hashlib.md5()
        with open(fname, 'rb') as f:
            while True:
                block = f.read(8192)
                if not block:
                    break
                md5_new.update(block)
        if md5_new.hexdigest() == md5.strip():
            self.log("Download skipped: %s (file exists, md5 verified)", name)
            return True
        return False

    async def _download_one(self, session, name, url, dest, md5):
        part = dest+".part"
        if md5:
            md5_new = hashlib.md5()
        try:
            async with self._sem, \
                       session.get(url, timeout=self._timeout) as resp:
                if not resp.status == 200:
                    self.log("Download failed: %s (error code %i)",
                             name, resp.status)
                    self.log("  URL: '%s'", url.strip())
                    return False
                size = int(resp.headers.get('content-length', 0))

                if os.path.exists(dest):
                    existing_size = os.path.getsize(dest)
                    if existing_size == size:
                        if md5:
                            self.log("Overwriting: %s (md5 failed)", name)
                        else:
                            self.log("Download skipped: %s (file exists)",
                                     name)
                            return True
                    else:
                        self.log("Overwriting: %s (size mismatch %i!=%i)",
                                 name, size, existing_size)

                try:
                    self._sum_bar.total += size
                except AttributeError:
                    pass
                with open(part, mode="wb") as out, \
                    tqdm(total=size,
                         unit='B', unit_scale=True, unit_divisor=1024,
                         desc=name, leave=False,
                         miniters=1, disable=not self._progress,
                         bar_format=self.make_bar_format(40, 7, rate=True)) as t:
                    while True:
                        block = await resp.content.read(self._block_size)
                        if not block:
                            break
                        out.write(block)
                        if md5:
                            md5_new.update(block)
                        t.update(len(block))
                        self._sum_bar.update(len(block))
            os.rename(part, dest)
            if md5:
                md5_hash = md5_new.hexdigest()
                if isinstance(md5, bool):
                    self.log("Download complete: %s (md5=%s)", name,
                             md5_hash.strip())
                elif md5.strip() == md5_hash:
                    self.log("Download complete: %s (md5 verified)", name)
                else:
                    self.error("Download failed: %s (md5 failed)", name)
                    return False
            return True
        except (asyncio.CancelledError, asyncio.TimeoutError):
            if os.path.exists(part):
                os.unlink(part)
            raise

    async def _run(self, urls: List[str], dest: str,
                   md5s: Optional[List[str]]=None) -> List[bool]:
        """Executes a download session

        Args:
          urls: List of URLs
          dest: Destination path
          md5s: Optional list of md5 checksums
        """
        if not md5s:
            md5s = [None]*len(urls)
        async with aiohttp.ClientSession() as session:
            if len(urls) == 0:
                # No need to show progress bar for just 1 file
                self.log("Downloading 1 file.")
                result = await asyncio.ensure_future(
                    self._download(session, urls[0], dest, md5s[0])
                )
                self.log("Finished download.")
            else:
                self.log("Downloading %i files.", len(urls))
                coros = [
                    asyncio.ensure_future(
                        self._download(session, url, dest, md5)
                    )
                    for url, md5 in zip(urls, md5s)
                ]
                with tqdm(
                    asyncio.as_completed(coros), total=len(coros),
                    unit="Files", desc="Total files:",
                    disable=not self._progress, leave=False,
                    bar_format=self.make_bar_format(20, 7, eta=True)
                ) as t, tqdm(
                    total=1,  # must be >0
                    unit="B", desc="Total bytes:",
                    unit_scale=True, unit_divisor=1024,
                    disable=not self._progress, leave=False, miniters=1,
                    bar_format=self.make_bar_format(20, 7, rate=True)
                ) as t2:
                    self._sum_bar = t2
                    result = [await coro for coro in t]
                self.log("Finished downloads.")
        return result

    def get(self, urls: Union[str, List[str]], dest: str,
            md5s: Optional[List[str]]=None) -> None:
        """Download a list of URLs

        Args:
          urls: List of URLs
          dest: Destination folder
          md5s: List of MD5 sums to check
        """

        urls = ensure_list(urls)
        if not urls:
            return True  # nothing to do
        if len(urls) > 1:
            if not os.path.exists(dest):
                os.makedirs(dest)

        try:
            task = asyncio.ensure_future(self._run(urls, dest, md5s))
            self.loop.run_until_complete(task)
        except KeyboardInterrupt:
            end = asyncio.gather(*asyncio.Task.all_tasks())
            end.cancel()
            try:
                self.loop.run_until_complete(end)
            except asyncio.CancelledError:
                pass
            raise

        return all(task.result())


class DownloadThread(object):
    def __init__(self):
        LOG.error("made downloader")
        self.loop = asyncio.new_event_loop()
        self.thread = threading.Thread(target=self.main)
        self.thread.start()
        atexit.register(self.terminate)

    def terminate(self):
        self.loop.call_soon_threadsafe(self.loop.stop)

    def main(self):
        LOG.error("here")
        asyncio.set_event_loop(self.loop)
        self.downloader = FileDownloader()
        self.loop.run_forever()

    def get(self, url, dest, md5):
        LOG.error("scheduling get %s", url)
        self.loop.call_soon_threadsafe(
            self.downloader.get(url, dest, md5)
        )


#DOWNLOADER = DownloadThread()

#def download(url, dest, md5=None):
#    LOG.error("called download %s", url)
#    DOWNLOADER.get(url, dest, md5)



import functools
import os
import re
import textwrap

from snakemake.io import Namedlist
from snakemake.utils import format as snake_format


def make_local_path(icfg, url: str):
    url_match = re.match("^(http|https|ftp|ftps)://", url)
    if url_match:
        return os.path.join(
            icfg.dir.downloads,
            url_match.group(1),
            url[url_match.end():]
        )
    return url


def is_fq(path):
    return isinstance(path, str) and (
        path.endswith(".fq.gz")
        or path.endswith(".fastq.gz")
        or path.endswith(".fq")
        or path.endswith(".fastq")
    )


def file_not_empty(fn):
    "Checks is a file is not empty, accounting for gz mininum size 20"
    if fn.endswith('gz'):
        return os.path.getsize(fn) > 20
    return os.path.getsize(fn) > 0


def filter_out_empty(*args):
    """
    Removes empty sets of files from input file lists.

    Takes a variable number of file lists of equal length and removes
    indices where any of the files is empty. Strings are converted to
    lists of length 1.

    Returns a generator tuple.

    Example:
    r1, r2 = filter_out_empty(input.r1, input.r2)
    """
    args = ([arg] if isinstance(arg, str) else arg
            for arg in args)
    return zip(*(t for t in zip(*args)
                 if all(map(file_not_empty, t))))


@functools.lru_cache()
def fasta_names(fasta_file):
    print("Calling fasta_names on {}".format(fasta_file))
    res = []
    with open(fasta_file, "r") as f:
        for line in f:
            if line[0] != ">":
                continue
            res += [line[1:].split(" ", 1)[0]]
    return res


def read_propfiles(files):
    if isinstance(files, str):
        files = [files]
    props = {}
    for file in files:
        with open(file, "r") as f:
            props.update({
                key: int(float(value))
                for line in f
                for key, value in [line.strip().split(maxsplit=1)]
            })
    return props


def glob_wildcards(pattern, files=None):
    """
    Glob the values of the wildcards by matching the given pattern to the
    filesystem.
    Returns a named tuple with a list of values for each wildcard.
    """
    from snakemake.io import _wildcard_regex, namedtuple, regex
    import regex as re

    pattern = os.path.normpath(pattern)
    first_wildcard = re.search("{[^{]", pattern)
    dirname = os.path.dirname(pattern[:first_wildcard.start(
    )]) if first_wildcard else os.path.dirname(pattern)
    if not dirname:
        dirname = "."

    names = [match.group('name')
             for match in _wildcard_regex.finditer(pattern)]
    Wildcards = namedtuple("Wildcards", names)
    wildcards = Wildcards(*[list() for name in names])

    pattern = regex(pattern)
    # work around partial matching bug in python regex module
    # by replacing matches for "\" with "[/\0]" (0x0 can't occur in filenames)
    pattern = re.sub('\\\\/', '[/\0]', pattern)
    cpattern = re.compile(pattern)

    def walker(dirname, pattern):
        """finds files/dirs matching `pattern` in `dirname`"""
        for dirpath, dirnames, filenames in os.walk(dirname):
            dirpath = os.path.normpath(dirpath)
            for f in filenames:
                if dirpath != ".":
                    f = os.path.join(dirpath, f)
                match = pattern.match(f)
                if match:
                    yield match
            for i in range(len(dirnames)-1, -1, -1):
                d = dirnames[i]
                if dirpath != ".":
                    d = os.path.join(dirpath, d)
                match = pattern.match(os.path.join(d, ""), partial=True)
                if not match:
                    del dirnames[i]
                    continue
                if match.partial:
                    continue
                yield match

    print("searching {}".format(pattern))
    if files is None:
        for match in walker(dirname, cpattern):
            for name, value in match.groupdict().items():
                getattr(wildcards, name).append(value)
    else:
        for f in files:
            match = re.match(cpattern, os.normpath(f))
            if match:
                for name, value in match.groupdict().items():
                    getattr(wildcards, name).append(value)
    print("searching {}: done".format(pattern))
    return wildcards


def activate_R():
    from rpy2.robjects import default_converter, conversion
    from rpy2 import robjects, rinterface

    @default_converter.py2ri.register(dict)
    def _1(obj):
        keys = list(obj.keys())
        res = rinterface.ListSexpVector([
            conversion.py2ri(obj[x])
            for x in keys
        ])
        res.do_slot_assign('names', rinterface.StrSexpVector(keys))
        return res

    @default_converter.py2ri.register(tuple)
    def _2(obj):
        return conversion.py2ri(list(obj))

    @default_converter.py2ri.register(list)
    def _3(obj):
        # return sequence_to_vector(obj)
        obj = rinterface.ListSexpVector([conversion.py2ri(x) for x in obj])
        return robjects.r.unlist(obj, recursive=False)


def R(code="", **kwargs):
    """Execute R code

    This function executes the R code given as a string. Additional arguments
    are injected into the R environment. The value of the last R statement
    is returned.

    The function requires rpy2 to be installed.

    Args:
        code (str): R code to be executed
        **kwargs (dict): variables to inject into R globalenv
    Yields:
        value of last R statement

    >>>  R("1*1", input=input)
    """
    try:
        import rpy2.robjects as robjects
        from rpy2.rlike.container import TaggedList
        from rpy2.rinterface import RNULLType
    except ImportError:
        raise ValueError(
            "Python 3 package rpy2 needs to be installed to use"
            "the R function.")

    activate_R()

    # translate Namedlists into rpy2's TaggedList to have named lists in R
    for key in kwargs:
        value = kwargs[key]
        if isinstance(value, Namedlist):
            kwargs[key] = TaggedList([y for x, y in value.allitems()],
                                     [x for x, y in value.allitems()])

    code = snake_format(textwrap.dedent(code), stepout=2)
    # wrap code in function to preserve clean global env and execute
    rval = robjects.r("function({}){{ {} }}"
                      "".format(",".join(kwargs), code))(**kwargs)

    # Reduce vectors of length 1 to scalar, as implicit in R.
    if isinstance(rval, RNULLType):
        rval = None
    if rval and len(rval) == 1:
        return rval[0]
    return rval


def Rmd(rmd, out, **kwargs):
    R("""
    library(rmarkdown)
    print(out)
    rmarkdown::render(rmd, params=paramx, output_file=out)
    """,
      rmd=rmd,
      out=os.path.abspath(out[0]), paramx=kwargs)



import logging
import os
import sys

import click

import tqdm

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

context_settings = {
    'help_option_names': ['-h', '--help']
}


class Group(click.Group):
    def command(self, *args, **kwargs):
        command = super().command(*args, context_settings=context_settings,
                                  **kwargs)

        def wrapper(f):
            return command(log_options(f))
        return wrapper


def command(*args, **kwargs):
    command = click.command(*args, context_settings=context_settings, **kwargs)

    def wrapper(f):
        return command(log_options(f))
    return wrapper


def group(*args, **kwargs):
    return command(*args, cls=Group, **kwargs)


def nohup(ctx, param, val):
    """
    Make YMP continue after the shell dies.

    - redirects stdout and stderr into pipes and sub process that won't
      die if it can't write to either anymore
    - closes stdin

    """
    if not val:
        return

    import signal
    from select import select
    from threading import Thread

    # ignore sighup
    signal.signal(signal.SIGHUP, signal.SIG_IGN)

    # sighup is actually only sent by bash if bash gets it itself, so not if
    # `exit` is called with ymp in the background. What usually kills that type
    # of process is trying to write to stdout or stderr. We need to catch that.

    # close stdin (don't need that anyway)
    sys.stdin.close()

    # redirect stdout and err into a pipe and save original target
    pipes = {}
    for fd in (sys.stdout, sys.stderr):
        # save original std fd
        saved = os.dup(fd.fileno())
        # create a pipe
        pipe = os.pipe()
        # overwrite std fd with one end of pipeo
        os.dup2(pipe[1], fd.fileno())
        # save other end and target std fd
        pipes[pipe[0]] = saved

    def watcher():
        while True:
            r, _, x = select(pipes.keys(), [], [], 60)
            for pipe in pipes:
                if pipe in r:
                    data = os.read(pipe, 4096)
                    try:
                        os.write(pipes[pipe], data)
                    except IOError:
                        pass

    t = Thread(target=watcher)
    t.daemon = True
    t.start()


nohup_option = click.option(
    "--nohup", "-N", is_flag=True,
    help="Don't die once the terminal goes away.",
    callback=nohup
)


class TqdmHandler(logging.StreamHandler):
    """Tqdm aware logging StreamHandler

    Passes all log writes through tqdm to allow progress bars
    and log messages to coexist without clobbering terminal
    """
    def emit(self, record):
        tqdm.tqdm.write(self.format(record))


class Log(object):
    """
    Set up Logging
    """
    # We could get snakemake's logging output like so:
    # slog = logging.getLogger("snakemake")
    # slog.parent = log
    #
    # Or use the log_handler parameter to snakemake to override
    # turning log events into messages. That would take some rewriting
    # of snakemake code though.
    #
    # There seems to be no good way of redirecting snakemakes logging
    # in a python-logging style way as it installs its own stream
    # handlers once control has passed to snakemake.
    def __init__(self):
        from coloredlogs import ColoredFormatter

        log = logging.getLogger("ymp")
        log.setLevel(logging.WARNING)
        log_handler = TqdmHandler()
        log_handler.setLevel(logging.DEBUG)  # no filtering here
        formatter = ColoredFormatter("YMP: %(message)s")
        log_handler.setFormatter(formatter)
        log.addHandler(log_handler)

        self.log = log

    def mod_level(self, n):
        new_level = self.log.getEffectiveLevel() + n*10
        clamped_level = max(logging.DEBUG, min(logging.CRITICAL, new_level))
        self.log.setLevel(clamped_level)

    @classmethod
    def verbose_option(cls, ctx, param, val):
        log = ctx.ensure_object(Log)
        if val:
            log.mod_level(-val)

    @classmethod
    def quiet_option(cls, ctx, param, val):
        log = ctx.ensure_object(Log)
        if val:
            log.mod_level(val)


verbose_option = click.option(
    "--verbose", "-v", count=True,
    help="Increase log verbosity",
    callback=Log.verbose_option,
    expose_value=False
)


quiet_option = click.option(
    "--quiet", "-q", count=True,
    help="Decrease log verbosity",
    callback=Log.quiet_option,
    expose_value=False
)


def enable_debug(ctx, param, val):
    if not val:
        return

    def excepthook(typ, val, tb):
        import traceback
        traceback.print_exception(typ, val, tb)
        import pdb
        pdb.pm()

    sys.excepthook = excepthook
    log.error("Dropping into PDB on uncaught exception...")


debug_option = click.option(
    "--python-debug", "-P", is_flag=True,
    help="Drop into debugger on uncaught exception",
    callback=enable_debug,
    expose_value=False
)


def log_options(f):
    f = verbose_option(f)
    f = quiet_option(f)
    f = debug_option(f)
    return f



import logging
import os
from hashlib import sha1
from typing import Dict, Optional
from collections.abc import Mapping, Sequence

from ymp.snakemake import make_rule
from ymp.util import make_local_path

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Archive(object):
    name = None
    hash = None
    tar = None
    dirname = None
    strip_components = None
    files = None

    def __init__(self, name, dirname, stage, tar, url, strip, files):
        self.name = name
        self.dirname = dirname
        self.stage = stage
        self.tar = tar
        self.url = url
        self.strip = strip
        self.files = files

        self.hash = sha1(self.tar.encode('utf-8')).hexdigest()[:8]
        self.prefix = os.path.join(self.dirname, "_unpacked_" + self.hash)

    def get_files(self):
        if isinstance(self.files, Sequence):
            return {self.stage + fn: os.path.join(self.prefix, fn)
                    for fn in self.files}
        elif isinstance(self.files, Mapping):
            return {self.stage + fn_ymp: os.path.join(self.prefix, fn_arch)
                    for fn_ymp, fn_arch in self.files.items()}
        else:
            raise Exception("unknown data type for reference.files")

    def make_unpack_rule(self, baserule: 'Rule'):
        docstr_tpl = """
        Unpacks {} archive:

        URL: {}

        Files:
        """

        item_tpl = """
        - {}
        """
        docstr = "\n".join([docstr_tpl.format(self.name, self.url)] +
                           [item_tpl.format(fn) for fn in self.files])
        return make_rule(
            name="unpack_{}_{}".format(self.name, self.hash),
            docstring=docstr,
            lineno=0,
            snakefile=__name__,
            parent=baserule,
            input=([], {'tar': self.tar}),
            output=([], {'files': list(self.get_files().values())}),
            params=([], {'strip': self.strip,
                         'prefix': self.prefix})
        )


class Reference(object):
    """
    Represents (remote) reference file/database configuration
    """
    ONEFILE = "ALL.contigs"

    def __init__(self, reference, cfg):
        self.name = "ref_" + reference
        import ymp
        cfgmgr = ymp.get_config()
        self.cfg = cfg
        self.files = {}
        self.archives = []
        self.dir = os.path.join(cfgmgr.dir.references,
                                reference)

        for rsc in cfg:
            if isinstance(rsc, str):
                rsc = {'url': rsc}
            type_name = rsc.get('type', 'fasta').lower()
            stage = rsc.get("stage", "") + "/"
            downloaded_path = make_local_path(cfgmgr, rsc['url'])
            if type_name == 'fasta':
                self.files[stage + 'ALL.fasta.gz'] = downloaded_path
            elif type_name == 'fastp':
                self.files[stage + 'ALL.fastp.gz'] = downloaded_path
            elif type_name == 'gtf':
                self.files[stage + 'ALL.gtf'] = downloaded_path
            elif type_name == 'snp':
                self.files[stage + 'ALL.snp'] = downloaded_path
            elif type_name == 'dir':
                archive = Archive(name=self.name,
                                  dirname=self.dir,
                                  stage=stage,
                                  tar=downloaded_path,
                                  url=rsc['url'],
                                  files=rsc['files'],
                                  strip=rsc.get('strip_components', 0))
                self.files.update(archive.get_files())
                self.archives.append(archive)
            else:
                log.debug("unknown type {} used in reference {}"
                          "".format(type_name, self.name))

        self.outputs = set(f.replace("ALL", "{sample}") for f in self.files)
        self.inputs = set()
        self.group = ["ALL"]

    @property
    def defined_in(self):
        return self.cfg.get_files()

    def match(self, name):
        return name == self.name

    def get_file(self, filename, stage):
        downloaded_path = self.files.get(stage + "/" + filename)
        if downloaded_path:
            return downloaded_path
        return ("YMP_FILE_NOT_FOUND__" +
                "No file {} in Reference {}"
                "".format(filename, self.name).replace(" ", "_"))

    def make_unpack_rules(self, baserule: 'Rule'):
        for archive in self.archives:
            yield archive.make_unpack_rule(baserule)

    def __str__(self):
        return os.path.join(self.dir, "ALL")



"""
This allows calling the YMP cli via ``python -m``

>>> python -m ymp.cli show references -v
"""

import sys
from ymp.cli import main

if __name__ == "__main__":
    sys.argv[0] = "ymp"
    main()



#!/usr/bin/env python3

import click
from ymp import blast
from ymp import gff


@click.command()
@click.argument('inf', metavar="input", type=click.File('r'))
@click.argument('out', metavar="output", type=click.File('w'))
def blast2gff(inf, out):
    blastfile = blast.reader(inf)
    gfffile = gff.writer(out)

    for hit in blastfile:
        assert (hit.send > hit.sstart) == (hit.sframe > 0)

        feature = gff.Feature(
            seqid=hit.sacc,
            source='BLAST',
            type='CDS',
            start=min(hit.sstart, hit.send),
            end=max(hit.sstart, hit.send),
            score=hit.evalue,
            strand='+' if hit.sframe > 0 else '-',
            phase='0',
            attributes="ID={}_{}_{}".format(hit.sacc, hit.sstart, hit.send)
        )
        gfffile.write(feature)


if __name__ == '__main__':
    # pylint does not get click decorators, disable warning:
    # pylint: disable=no-value-for-parameter
    blast2gff()



"""
Implements simple reader and writer for GFF (general feature format) files.

Unfinished

 - only supports one version, GFF 3.2.3.
 - no escaping
"""


from collections import namedtuple

_FIELDS = [
    'seqid',   # the sequence id
    'source',  # the tool
    'type',    # CDS, rRNA, ...
    'start',
    'end',
    'score',
    'strand',  # + / -
    'phase',
    'attributes'
]

Feature = namedtuple("Feature", _FIELDS)

_ATTRIBUTES = [
    'ID',      # unique ID per GFF, use multiple for discontinuous features
    'Name',    # display name for user
    'Alias',   # secondary name, e.g. accession
    'Parent',  # indicate "partof" relationship
    'Target',  # alignment target to 'target_id start end [strand]'
    'Gap',     # alignment in CIGAR format
    'Derives_From',  # temporal relationship
    'Note',    # free text
    'Dbxref',  # database cross reference
    'Ontology_term',
    'Is_circular',  # if feature is circular
    # note
    # product
    # partial
    # start_type
    # rbs_motif
    # rbs_spacer
    # gc_cont
    # conf
    # score
    # <x>score
]

Attributes = namedtuple("Attributes", _ATTRIBUTES)


class reader(object):
    def __init__(self, fileobj):
        self.fileobj = fileobj

    def __iter__(self):
        for line in self.fileobj:
            if isinstance(line, bytes):
                line = line.decode('ascii')
            if line[0] == "#":
                continue
            f = line.strip().split('\t')
            f[3] = int(f[3])
            f[4] = int(f[4])
            f[-1] = dict([
                tuple(item.split('='))
                for item in f[-1].strip(';').split(';')
            ])
            yield Feature(*f)


class writer(object):
    def __init__(self, fileobj):
        self.fileobj = fileobj
        self.fileobj.write("##gff-version 3\n")

    def write(self, feature):
        self.fileobj.write("\t".join(map(str, feature))+"\n")



import sys, csv, re
from argparse import ArgumentParser
import fileinput

class emirge_info:
    def __init__(self, line):
        ## C5-14863|JF198678.1.1366;Prior=0.000387;Length=1357;NormPrior=0.000383;size=383
        ## <sample>-<id>|<acc>;(<key>=<value>)*[;]
        field_list = line.split(";");
        self.sid, self.acc = field_list[0].split("|")
        self.sample, self.id, _ = re.split('-([0-9]*)$', self.sid)

        for kv in field_list[1:]:
            try:
                (key, value) = kv.split("=")
                try:
                    self.__dict__[key] = float(value)
                except ValueError:
                    self.__dict__[key] = value
            except:
                pass

    def __str__(self):
        return "sample=%s id=%s acc=%s prior=%f" % (self.sample, self.id, self.acc, self.Prior)


class MapfileParser(object):
    def __init__(self, minid=0):
        self.samples = set()
        self.by_centroid_sums={}
        self.by_sample_sums={}
        self.minid=0

    def read(self, mapfiles):
        mapfile = fileinput.input(mapfiles)
        try:
            for line in mapfile:
                items = line.split("\t")
                src = emirge_info(items[0])
                centroid = emirge_info(items[1])
                id = float(items[2])
                if id < self.minid:
                    continue

                self.samples.add(src.sample)
            
                if src.sample in self.by_sample_sums:
                    self.by_sample_sums[src.sample] += src.NormPrior
                else:
                    self.by_sample_sums[src.sample] = src.NormPrior

                if not centroid.sid in self.by_centroid_sums:
                    self.by_centroid_sums[centroid.sid] = {"centroid": centroid.sid}

                if src.sample in self.by_centroid_sums[centroid.sid]:
                    self.by_centroid_sums[centroid.sid][src.sample] += src.NormPrior
                else:
                    self.by_centroid_sums[centroid.sid][src.sample] = src.NormPrior
        finally:
            mapfile.close()

    def write(self, outfile):
        if outfile == "-":
            outfile = sys.stdout
        else:
            outfile = open(outfile, "w")

        try:
            writer = csv.DictWriter(outfile, fieldnames = ["centroid"] + sorted(self.samples))
            writer.writeheader()
            for key in sorted(self.by_centroid_sums.keys()):
                writer.writerow(self.by_centroid_sums[key])
        finally:
            if outfile != sys.stdout:
                outfile.close()


def main():
    parser = ArgumentParser()
    parser.add_argument("-o","--output", default="-", metavar="file")
    parser.add_argument("--id", type=float, default=0.0)
    parser.add_argument("input", nargs="*", default="-")
    args = parser.parse_args()

    parser = MapfileParser(args.id)
    parser.read(args.input)
    parser.write(args.output)
    

if __name__ == '__main__':
    main()



import logging
import os

import click

import pytest

import ymp
cfg = ymp.get_config()

log = logging.getLogger(__name__)  # pylint: disable=invalid-name


def test_submit_no_profile(invoker):
    "Running submit without profile should raise UsageError"
    with pytest.raises(click.UsageError):
        invoker.call("submit")


def test_submit_profile_cfg(invoker, saved_tmpdir):
    "Test profile set from config"
    with open("ymp.yml", "w") as ymp_yml:
        ymp_yml.write("cluster:\n  profile: dummy")
    invoker.call("submit", cfg.dir.reports)
    assert os.path.isdir(cfg.dir.reports)


# - don't test profiles that have no command set on them (default profile)
# - sort profiles so we can test in parallel reliably
profiles = sorted((name, profile)
                  for name, profile in cfg.cluster.profiles.items()
                  if profile.get('command'))


@pytest.mark.parametrize(
    "mock_cmd,prof_name,prof_cmd",
    [((profile.command.split()[0], '#!/bin/bash\nexec "$@"\n'),
      name, profile.command)
     for name, profile in profiles],
    ids=[name for name, profile in profiles],
    indirect=['mock_cmd'])
def test_submit_profiles(invoker, mock_cmd, prof_name, prof_cmd):
    invoker.call("submit",
                 "--profile", prof_name,
                 "--command", prof_cmd,
                 cfg.dir.reports)
    assert os.path.isdir(cfg.dir.reports)


def test_show(invoker, saved_tmpdir):
    "Test parts of show"
    with open("ymp.yml", "w") as cfg:
        cfg.write("conda:\n  testme: [X1,X2]")

    invoker.call("show")
    res = invoker.call("show", "pairnames")
    assert res.output.strip() == '- R1\n- R2'
    res = invoker.call("show", "pairnames[1]")
    assert res.output.strip() == 'R2'
    res = invoker.call("show", "cluster.profiles.default.drmaa")
    assert res.output.strip() == 'False'

    res = invoker.call("show", "conda.testme")
    assert res.output.strip() == '- X1\n- X2'


def test_stage_list(invoker):
    "List all stages"
    res = invoker.call("stage", "list")
    assert "\ncheck " in res.output

    with pytest.raises(click.UsageError):
        res = invoker.call("stage", "list", "-s", "-l")

    res = invoker.call("stage", "list", "does_not_exist")
    assert res.output == ""

    res = invoker.call("stage", "list", "ch?ck", "-s")
    assert res.output.strip() == "check"

    res = invoker.call("stage", "list", "ch?ck", "-l")
    assert res.output.startswith("check")
    assert res.output.count("\n") > 3

    res = invoker.call("stage", "list", "ch?ck", "-c")
    assert res.output.startswith("check")
    assert "test.rules:" in res.output
    assert res.output.count("\n") == 3


def test_func_get_envs():
    "Test env cli helper function get_envs"
    from ymp.cli.env import get_envs
    cfg.unload()

    envs = get_envs()
    log.debug("envs found: %s", envs)
    assert 'bbmap' in envs
    assert 'bmtagger' in envs

    envs = get_envs('bbmap')
    assert len(envs) == 1

    envs = get_envs(['bbmap', 'bmtagger'])
    assert len(envs) == 2

    envs = get_envs(['bb?ap', 'bmtagger*'])
    assert len(envs) == 2


def test_env_list(invoker):
    """Test listing environments

    - w/o args
    - reverse sorted
    - sorted by hash
    """
    res = invoker.call("env", "list")
    lines = res.output.splitlines()
    assert len(lines) > 2
    assert lines[0].startswith("name"), "first row should start with name"
    assert all(lines[i].upper() <= lines[i+1].upper()
               for i in range(2, len(lines)-1)), \
        f"output should be sorted: {lines}"

    res = invoker.call("env", "list", "-r")
    lines = res.output.splitlines()
    assert all(lines[i].upper() >= lines[i+1].upper()
               for i in range(1, len(lines)-1)), \
        f"output should be sorted reverse:\n{lines}"

    res = invoker.call("env", "list", "-s", "hash")
    lines = res.output.splitlines()
    hash_col = lines[0].split().index("hash")
    hashes = [line.split()[hash_col] for line in lines]
    assert all(hashes[i] <= hashes[i+1]
               for i in range(1, len(lines)-1)), \
        f"output should be sorted by hash:\n{hashes[1:]}"


def test_env_prepare(invoker, demo_dir, mock_conda, mock_downloader):
    """Test passing through to snakemake prepare"""
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")
    res = invoker.call("env", "list", "bbmap")
    lines = res.output.splitlines()
    col = lines[0].index("installed")
    assert lines[1][col:col+len("False")] == "False"
    invoker.initialized = False
    res = invoker.call("env", "prepare", "toy.trim_bbmap")

    res = invoker.call("env", "list", "bbmap")
    lines2 = res.output.splitlines()
    col = lines2[0].index("installed")
    assert lines2[1][col:col+len("True")] == "True", "\n".join(lines + lines2)

    conda_cmd = mock_conda.calls[-1]
    assert "conda create" in conda_cmd
    assert "/bbmap-" in conda_cmd


def test_env_install(invoker, demo_dir, mock_conda, mock_downloader):
    """Test installing environments"""
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")

    # basic
    res = invoker.call("env", "install", "bbmap")
    assert "Creating 1 environments" in res.output
    assert "'bbmap'" in res.output
    assert "--prefix "+str(demo_dir) in mock_conda.calls[0]
    assert len(mock_conda.calls) == 1

    # no double install
    res = invoker.call("env", "install", "bbmap")
    assert len(mock_conda.calls) == 1

    # remove bbmap env
    res = invoker.call("env", "remove", "bbmap")

    # multiple, globbing
    res = invoker.call("env", "install", "bb?ap", "bbma*")
    assert "Creating 1 environments" in res.output
    assert "'bbmap'" in res.output
    assert "--prefix "+str(demo_dir) in mock_conda.calls[1]
    assert len(mock_conda.calls) == 2

    # dynamic env
    res = invoker.call("env", "install", "sickle")
    assert "Creating 1 environments" in res.output
    assert "'sickle'" in res.output
    assert "--prefix "+str(demo_dir) in mock_conda.calls[2]


def test_env_update(invoker, demo_dir, mock_conda, mock_downloader):
    """Test updating environments"""
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")
    # basic
    res = invoker.call("env", "update", "bbmap")
    assert "Updating 1 environments" in res.output
    assert "'bbmap'" in res.output
    assert "conda create" in mock_conda.calls[0]
    assert "conda env update" in mock_conda.calls[1]


def test_env_export(invoker, demo_dir, mock_conda, mock_downloader):
    """Test exporting environments"""
    # install envs locally
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")

    # skip and create are mutually exclusive
    with pytest.raises(click.UsageError) as exc:
        invoker.call("env", "export", "-sc", "bbmap")
    assert exc.match("mutually exclusive")

    # fail if trying to export uninstalled env
    with pytest.raises(click.UsageError) as exc:
        res = invoker.call("env", "export", "bbmap")
    assert exc.match("uninstalled")

    # exporting nothing
    res = invoker.call("env", "export", "-qs", "bbmap")
    assert res.output == ""

    # creating bbmap.yml
    res = invoker.call("env", "export", "-d", ".", "bbmap", "-c")
    assert "Exporting" in res.output

    # fail, file exists
    with pytest.raises(click.UsageError) as exc:
        res = invoker.call("env", "export", "-d", ".", "bbmap")
    assert exc.match("exists")

    # allow overwrite
    res = invoker.call("env", "export", "-d", ".", "bbmap", "-f")
    assert "Exporting" in res.output

    # try txt format and export to file
    res = invoker.call("env", "export", "-d", "bbmap.txt", "bbmap")
    assert "conda list" in mock_conda.calls[-1]

    # export multiple, fail
    with pytest.raises(click.UsageError) as exc:
        res = invoker.call("env", "export", "-cd", ".", "bbmap", "sambamba")
    assert exc.match("exists")

    # export multiple
    res = invoker.call("env", "export", "-fcd", ".", "bbmap", "sambamba")
    assert "Exporting 2 " in res.output

    # export multiple to stdout
    res = invoker.call("env", "export", "-q", "bbmap", "sambamba")
    names = [line[6:] for line in res.output.splitlines()
             if line.startswith("name: ")]
    assert sorted(names) == ["bbmap", "sambamba"]

    # export no matching patterns
    res = invoker.call("env", "export", "does_not_match_anything")
    assert "Nothing to export" in res.output

    # export everything installed
    res = invoker.call("env", "export", "-s")
    names = [line[6:] for line in res.output.splitlines()
             if line.startswith("name: ")]
    assert sorted(names) == ["bbmap", "sambamba"]


def test_env_clean(invoker, demo_dir, mock_conda):
    """Test cleaning environments"""
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")


def test_env_activate(invoker, demo_dir, mock_conda, mock_downloader):
    """Test activating an environment"""
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")
    res = invoker.call("env", "activate", "bbmap")
    assert str(demo_dir) in res.output


def test_env_run(invoker, demo_dir, mock_conda, mock_downloader, capfd):
    with open("ymp.yml", "a") as f:
        f.write("directories:\n conda_prefix: '.'")

    with pytest.raises(click.UsageError) as exc:
        res = invoker.call("env", "run", "bbmapx", "bbmap.sh")
    assert exc.value.message == "Environment bbmapx unknown"

    with pytest.raises(click.UsageError) as exc:
        res = invoker.call("env", "run", "*", "bbmap.sh")
    assert exc.value.message.startswith("Multiple environments match")

    res = invoker.call("env", "run", "bbmap", "true")
    assert res.exit_code == 0
    cap = capfd.readouterr()
    assert "Not a conda environment" in cap.err


@pytest.mark.parametrize(
    "words,result",
    [
        ["ymp make", [
            "toy", "toy.", "mpic", "mpic.", 0
        ]],
        ["ymp make t", [
            "toy", "toy.", 0
        ]],
        ["ymp make toy.", [
            "toy.assemble_", "toy.trim_"
        ]],
        ["ymp make toy.assemble_", [
            "toy.assemble_megahit",
            "toy.assemble_megahit."
         ]],
        ["ymp make toy.assemble_megahit.", [
            "toy.assemble_megahit.trim_",
            "toy.assemble_megahit.map_"
        ]],
        ["ymp make toy.assemble_megahit.map_", [
            "toy.assemble_megahit.map_bbmap",
        ]],
        ["ymp make toy.map_bowtie2.", [
            0
        ]],
        ["ymp make toy.group_", [
            "toy.group_name", "toy.group_Subject",
            "toy.group_name.", "toy.group_Subject.",
            10
        ]],
    ]
)
def test_completion(invoker, demo_dir, capfd, envvar, words, result):
    import subprocess as sp
    envvar('YMP_DEBUG_EXPAND', 'stderr')
    envvar('_YMP_COMPLETE', 'complete-bash')
    envvar('COMP_CWORD', '2')
    envvar('COMP_WORDS', words)
    sp.run(["python", "-m", "ymp"])
    cap = capfd.readouterr()
    expanded = set(cap.out.split())

    for val in result:
        if isinstance(val, str):
            expanded.remove(val)
        else:
            assert len(expanded) == val



"""
This module contains helper functions.

Not all of these are currently in use
"""

from collections import Mapping, OrderedDict


class OrderedDictMaker(object):
    """
    odict creates OrderedDict objects in a dict-literal like syntax

    >>>  my_ordered_dict = odict[
    >>>    'key': 'value'
    >>>  ]

    Implementation:
    odict uses the python slice syntax which is similar to dict literals.
    The [] operator is implemented by overriding __getitem__. Slices
    passed to the operator as ``object[start1:stop1:step1, start2:...]``,
    are passed to the implementation as a list of objects with start, stop
    and step members. odict simply creates an OrderedDictionary by iterating
    over that list.
    """

    def __getitem__(self, keys):
        if isinstance(keys, slice):
            return OrderedDict([(keys.start, keys.stop)])
        return OrderedDict([(slice.start, slice.stop) for slice in keys])


odict = OrderedDictMaker()  # pylint: disable=invalid-name


def update_dict(dst, src):
    """Recursively update dictionary ``dst`` with ``src``

    - Treats a `list` as atomic, replacing it with new list.
    - Dictionaries are overwritten by item
    - None is replaced by empty dict if necessary
    """
    if src is None:
        return dst
    for key, val in src.items():
        if isinstance(val, Mapping):
            dst_sub = dst.get(key, {})
            if (dst_sub) is None:
                dst_sub = {}
            tmp = update_dict(dst_sub, val)
            dst[key] = tmp
        else:
            dst[key] = src[key]
    return dst



import logging

import pytest

from ymp.stage import Stage

log = logging.getLogger(__name__)


def get_targets():
    stages = Stage.get_registry()

    ympmakedoc = ">>> ymp make "
    stage_testcount = {stage: 0 for stage in stages}
    targets = []
    for stage in stages.values():
        for line in stage.docstring.splitlines():
            if line.strip().startswith(ympmakedoc):
                stage_testcount[stage.name] += 1
                target = line.strip()[len(ympmakedoc):]
                targets.append(target)
    return targets


targets = get_targets()


@pytest.mark.parametrize("targetx", targets)
def test_stage_dryrun(invoker, targetx):
    invoker.call("init", "demo")
    invoker.call("make", "-n", targetx)


@pytest.mark.runs_tool
@pytest.mark.parametrize("targetx", targets)
def test_stage_run(invoker, targetx):
    invoker.call("init", "demo")
    invoker.call("make", targetx)



"""
Testing extensions to snakemake

TODO:
 - recursive parameter expansion
   X detect loop
   - numbered parameters
   - named parameters
   - named parameter sets
   - functions
 - ymp config expansion
   - general
   - localized
   - grouping
 - defaults
"""

import logging

import pytest

log = logging.getLogger(__name__)


@pytest.mark.parametrize("project", ["snakemake_circle"], indirect=True)
def test_snakemake_failure(project_dir, invoker):
    "These are expected to fail"
    res = invoker.call_raises("make", "test")
    msg = str(res.output)
    assert "Circular reference in" in msg


@pytest.mark.parametrize("project", ["snakemake_plain", "snakemake_function"],
                         indirect=True)
def test_snakemake(project_dir, invoker):
    "These should work"
    res = invoker.call("make", "test")
    assert res.exit_code == 0
